"""Transformer module for the transformation of lark trees."""

import logging
import re
import typing
from collections import Counter, OrderedDict, defaultdict, namedtuple
from typing import DefaultDict, Dict, Generator, List, Set

from lark import Transformer
from lark.lexer import Token
from lark.tree import Tree

from ebel.cache import logger
from ebel.constants import (ALLOWED_TYPES, FILE, GRAMMAR_START_ANNO,
                            GRAMMAR_START_NS, LIST, PATTERN, URL)
from ebel.errors import (NotDownloadedFromUrl, NotInAnnotationList,
                         NotInAnnotationPattern, NotInAnnotationUrl,
                         NotInNamespaceList, NotInNamespacePattern,
                         NotInNamespaceUrl, WithoutDefinedAnnotation,
                         WithoutDefinedNamespace, _Error)
from ebel.manager.models import Annotation as AnnotationModel
from ebel.manager.models import AnnotationEntry, AnnotationManager
from ebel.manager.models import Namespace as NamespaceModel
from ebel.manager.models import NamespaceEntry, NamespaceManager, reset_tables
from ebel.tools import BelRdb
from ebel.warnings import AlsoUsedInOtherNamespace, _Warning

log = logging.getLogger(__name__)


def get_token_dict2(token_tree_list: list) -> dict:
    """Create a dict out of a mixed list of lark.lexer.Token and tuple('name_of_rule',list(lark.lexer.Token,...)).

    Parameters
    ----------
    token_tree_list : list
        Description of parameter `token_tree_list`.

    Returns
    -------
    type
        Description of returned object.

    """
    ret_dict = defaultdict(list)

    for item in token_tree_list:
        if isinstance(item, Token):
            ret_dict[item.type].append(item.value)

        elif isinstance(item, tuple):
            ret_dict2 = defaultdict(list)

            for item2 in item[1]:
                if isinstance(item2, Token):
                    ret_dict2[item2.type].append(item2.value)

            ret_dict[item[0]] = dict(ret_dict2)
        elif isinstance(item, Tree):
            for item3 in item.children:
                ret_dict[item3.type].append(item3.value)

    return dict(ret_dict)


class _BelTransformer(Transformer):
    """Transformer optimized for JSON output.

    Vertices
    --------
        are tuples with 2 elements
        [0] := vertex name
        [1] := values

    Properties of vertices
    ----------------------
        are in OrderedDict's
        - properties with 1 value => namedtuple
        - properties with * values => List[namedtuple]
    """

    exclude_token_types = ("OB", "CB", "QM", "COLON", "COMMA", "OCB", "CCB")

    nt_frag = namedtuple("fragment", ("range", "descriptor"))
    nt_nn = namedtuple("nn", ("namespace", "name"))
    nt_var = namedtuple("variant", ("hgvs",))
    nt_ma = namedtuple("ma", ("namespace", "name", "default"))
    nt_pmod = namedtuple("protein_modification", ("namespace", "name", "type", "amino_acid", "position"))
    nt_support = namedtuple("support", ("text",))
    nt_citation = namedtuple("citation", ("type", "title", "ref", "pub_date", "author_list", "comment"))
    nt_type = namedtuple("function", ("type", "name"))
    nt_gmod = namedtuple("gene_modification", ("namespace", "name"))
    named_tuples = (
        nt_frag,
        nt_nn,
        nt_var,
        nt_pmod,
        nt_support,
        nt_citation,
        nt_type,
        nt_ma,
        nt_gmod,
    )

    def __init__(self, force_new_db=False):
        Transformer.__init__(self)
        self.cache = _BelScript(force_new_db=force_new_db)
        self.citations = []

    @staticmethod
    def script(n):
        return n

    @staticmethod
    def statements_and_sets(n):
        return {"statements_and_sets": n}

    def abundance(self, n):
        """Return abundance."""
        return [self.nt_type("abundance", "abundance"), n]

    def gene(self, n):
        """Return gene."""
        return [self.nt_type("abundance", "gene"), n]

    def micro_rna(self, n):
        """Return micro_rna."""
        return [self.nt_type("abundance", "micro_rna"), n]

    def rna(self, n):
        """Return rna."""
        return [self.nt_type("abundance", "rna"), n]

    def protein(self, n):
        """Return protein."""
        return [self.nt_type("abundance", "protein"), n]

    def population(self, n):
        """Return protein."""
        return [self.nt_type("abundance", "population"), n]

    def composite(self, n):
        """Return composite."""
        return [self.nt_type("list", "composite"), sorted(n)]

    @staticmethod
    def definitions(n):
        """Return definitions of namespace and annotations."""
        return {"definitions": n}

    def sec(self, n):
        """Return cellSecretion."""
        return [self.nt_type("transformation", "cell_secretion"), [n[0]]]

    def _ns_anno_props(self, n):
        """Return properties of namespace or annotation."""
        token_names = (
            "KEYWORD",
            ("URL", "PATTERN", "LIST"),
            ("URL_DEF", "PATTERN_DEF", "FILE_PATH"),
        )

        keyword, type_, value = self._get_values(n, token_names)

        value_list = []
        if type_ == "LIST":
            in_list = self._get_tree(n, "in_list")
            value_list = self._get_all_values_by_name(in_list, "ENTRY")

        props = OrderedDict(
            [
                ("keyword", keyword),
                ("type", type_),
                ("value", value),
                ("value_list", sorted(value_list)),
            ]
        )
        return props

    def namespace(self, n):
        """Return namespace vertex."""
        token_dict = get_token_dict2(n)

        keyword = token_dict["KEYWORD"][0]
        namespace_type, value = None, None

        if "LIST" in token_dict:
            value = token_dict["ENTRY"]
            namespace_type = LIST

        elif "URL_DEF" in token_dict:
            value = token_dict["URL_DEF"][0]
            namespace_type = URL

        elif "FILE_PATH" in token_dict:
            value = token_dict["FILE_PATH"][0]
            namespace_type = FILE

        elif "PATTERN_DEF" in token_dict:
            value = token_dict["PATTERN_DEF"][0]
            namespace_type = PATTERN

        self.cache.set_namespace_definition(namespace_type, keyword, value)

        props = self._ns_anno_props(n)
        return {"namespace": props}

    def annotation(self, n):
        """Return namespace vertex."""
        token_dict = get_token_dict2(n)

        keyword = token_dict["KEYWORD"][0]
        annotation_type, value = None, None

        if "LIST" in token_dict:
            value = token_dict["ENTRY"]
            annotation_type = LIST

        elif "URL_DEF" in token_dict:
            value = token_dict["URL_DEF"][0]
            annotation_type = URL

        elif "FILE_PATH" in token_dict:
            value = token_dict["FILE_PATH"][0]
            annotation_type = FILE

        elif "PATTERN_DEF" in token_dict:
            value = token_dict["PATTERN_DEF"][0]
            annotation_type = PATTERN

        self.cache.set_annotation_definition(annotation_type, keyword, value)

        props = self._ns_anno_props(n)
        return {"annotation": props}

    @staticmethod
    def document(n):
        """Return Document vertex."""
        return {"document": OrderedDict(sorted(n))}

    def _doc_prop(self, name, n):
        return (name, self._get_values(n, (("STRING_IN_QUOTES", "WORD"),))[0])

    def document_name(self, n):
        """Return document property as tuple."""
        return self._doc_prop("name", n)

    def document_description(self, n):
        """Return document property as tuple."""
        return self._doc_prop("description", n)

    def document_version(self, n):
        """Return document property as tuple."""
        return self._doc_prop("version", n)

    def document_authors(self, n):
        """Return document property as tuple."""
        return self._doc_prop("authors", n)

    def document_contact_info(self, n):
        """Return document property as tuple."""
        return self._doc_prop("contact_info", n)

    def document_copyright(self, n):
        """Return document property as tuple."""
        return self._doc_prop("copyright", n)

    def document_licences(self, n):
        """Return document property as tuple."""
        return self._doc_prop("licences", n)

    def document_keywords(self, n):
        """Return document property as tuple."""
        return self._doc_prop("keywords", n)

    def deg(self, n):
        """Return degradation."""
        return [self.nt_type("transformation", "degradation"), [n[0]]]

    def complex_abundance(self, n):
        """Return complex as abundance."""
        return [self.nt_type("abundance", "complex"), [n[0]]]

    def complex_list(self, n):
        """Return complex as list."""
        return [self.nt_type("list", "complex"), n]

    @staticmethod
    def list_complex(n):
        """Return abundance list of complex."""
        return sorted(n)

    def list(self, n):
        """Return list."""
        function_type = self._format_sub_obj([self.nt_type("list", "list")])[0]
        return {"object": [function_type, [self._format_sub_obj(x) for x in sorted(n)]]}

    def act2(self, n):
        """Return activity."""
        abundance = self._get_tree(n, "act_abundance").children[0]
        ma = self._get_dict_value(n, "ma")
        if ma:
            return [self.nt_type("process", "activity"), abundance, ma._asdict()]
        else:
            return [self.nt_type("process", "activity"), abundance]

    def act(self, n):
        """Return activity."""
        return [self.nt_type("process", "activity"), n]

    def tloc(self, n):
        """Return tloc."""
        return [self.nt_type("transformation", "translocation"), n]

    @staticmethod
    def _get_rel(n):
        return {"relation": n[0].data}

    @staticmethod
    def nested_relation(n):
        """Return nested relation."""
        return {"nested_relation": n[0]}

    def relation_with_list(self, n):
        """Return relation."""
        return self._get_rel(n)

    def relation_basic(self, n):
        """Return relation."""
        return self._get_rel(n)

    def rel_with_list_object(self, n):
        """Return relation."""
        return self._get_rel(n)

    def biomarker_relation(self, n):
        """Return relation."""
        return self._get_rel(n)

    def process_relation(self, n):
        """Return relation."""
        return self._get_rel(n)

    def analogous_relation(self, n):
        """Return relation."""
        return self._get_rel(n)

    def orthologous_relation(self, n):
        """Return orthologous relation."""
        return self._get_rel(n)

    def additional_relations(self, n):
        """Return relation."""
        return self._get_rel(n)

    def transc_relation(self, n):
        """Return relation."""
        return self._get_rel(n)

    def transl_relation(self, n):
        """Return relation."""
        return self._get_rel(n)

    @staticmethod
    def has_members(n):
        """Return relation."""
        return {"relation": "has_members"}

    @staticmethod
    def statement(n: list):
        """Return statement."""
        return {"statement": n}

    def subject(self, n):
        """Return subject."""
        return {"subject": self._format_sub_obj(n[0])}

    def nested_subject(self, n):
        """Return nested subject."""
        return {"nested_subject": self._format_sub_obj(n[0])}

    def nested_object(self, n):
        """Return nested subject."""
        return {"nested_object": self._format_sub_obj(n[0])}

    @staticmethod
    def sets(n):
        """Return sets."""
        return {"sets": n}

    @staticmethod
    def evidence(n):
        """Return support.

        support formily known as evidence, supprtingText
        """
        return {"evidence": n[0].value}

    def citation(self, n):
        """Return citation."""
        tv_dict = {x.type: x.value for x in n if isinstance(x, Token)}
        c_title, c_pubdate, c_author_list, c_comment = "", "", "", ""
        c_type = tv_dict["C_TYPE"]

        if "ONLY_2_PARAMETERS" in tv_dict:
            c_ref = tv_dict["C_PARAM2"]
        else:
            c_title = tv_dict["C_PARAM2"]
            c_ref = tv_dict["C_PARAM3"]
            c_pubdate = tv_dict["C_PUBDATE"] if "C_PUBDATE" in tv_dict else ""
            c_author_list = tv_dict["C_AUTHORLIST"] if "C_AUTHORLIST" in tv_dict else ""
            c_comment = tv_dict["C_COMMENT"] if "C_COMMENT" in tv_dict else ""

        nt = self.nt_citation(
            type=c_type,
            title=c_title,
            ref=c_ref,
            pub_date=c_pubdate,
            author_list=c_author_list,
            comment=c_comment,
        )

        self.citations.append({"citation_type": c_type, "citation_id": c_ref})

        return {"citation": nt._asdict()}

    def statement_group(self, n):
        """Return statement group."""
        return {"statement_group": self._get_value(n, "GROUP_NAME")}

    def set_annotation(self, n):
        """Return annotation sets."""
        entries = []
        annotation_key = ""
        for tkn in n:
            if tkn.type == "KEYWORD":
                annotation_key = tkn.value
            elif tkn.type == "ANNO_SET_ENTRY":
                entries.append(tkn)

        for entry in entries:
            self.cache.set_annotation_entry(annotation=annotation_key, entry=entry.value, token=entry)

        keyword = self._get_value(n, "KEYWORD")
        entries = self._get_all_values_by_name(n, "ANNO_SET_ENTRY")
        return {"set": {keyword: entries}}

    def unset(self, n):
        """Return unsets."""
        keywords = self._get_all_values_by_name(n, "ANNO_KEYWORD")
        return {"unset": keywords}

    def object(self, n):
        """Return object."""
        return {"object": self._format_sub_obj(n[0])}

    def molec_process(self, n):
        """Return molecular process."""
        return self.object(n)

    def act_or_abundance(self, n):
        """Return activity or abundance."""
        return {"act_or_abundance": self._format_sub_obj(n[0])}

    def transformation(self, n):
        """Return transformation."""
        return {"transformation": self._format_sub_obj(n[0])}

    def pat(self, n):
        """Return object."""
        return self.subject(n)

    def transc_subject(self, n):
        return self.subject(n)

    def transc_object(self, n):
        return self.object(n)

    def transl_subject(self, n):
        return self.subject(n)

    def transl_object(self, n):
        return self.object(n)

    def ortho_subject(self, n):
        """Return orthologous gene subject."""
        return self.subject(n)

    def ortho_object(self, n):
        """Return orthologous gene object."""
        return self.object(n)

    def surf(self, n):
        return [self.nt_type("transformation", "cell_surface_expression"), n]

    def bp(self, n):
        """Return biological process."""
        return [self.nt_type("process", "biological_process"), [n[0]]]

    def path(self, n):
        """Return path."""
        return [self.nt_type("process", "pathology"), [n[0]]]

    def _format_sub_obj(self, obj):
        """Change namedtuples to tuple(name_of_namedtuple,dictionary)."""
        if isinstance(obj, (list, tuple)):
            for i in range(len(obj)):
                if not isinstance(obj[i], self.named_tuples) and isinstance(obj[i], (list, tuple)):
                    self._format_sub_obj(obj[i])
                elif isinstance(obj[i], self.named_tuples):
                    function_type = obj[i].__class__.__name__
                    if function_type == "function":
                        obj[i] = {function_type: obj[i]._asdict()}
                    else:
                        obj[i] = obj[i]._asdict()
        return obj

    def protein_changes(self, n):
        """Create dict of protein changes."""
        return self.changes(n)

    @staticmethod
    def statement_comment(n):
        return {"statement_comment": " ".join([x.value for x in n]).strip()}

    @staticmethod
    def changes(n):
        """Create dict of abundance changes."""
        return [list((x[0], [x[1]])) for x in sorted(list(set(n)))]

    @staticmethod
    def gene_changes(n):
        """Return gene abundance changes; Only valid in BEL 2.1."""
        return [list((x[0], [x[1]])) for x in sorted(list(set(n)))]

    def gmod(self, n):
        """Return gmod; Only valid in BEL 2.1."""
        nn = [x for x in n if x.__class__.__name__ == "nn"]
        namespace = "TestNS"
        name = "TestName"

        if nn:
            namespace = nn[0].namespace
            name = nn[0].name
        nt = self.nt_gmod(namespace=namespace, name=name)

        return self.nt_type("modifier", "gmod"), nt

    def ma(self, n):
        """Transform tokens to dict(function_name: namedtuple)."""
        namespace, name, default = (None,) * 3
        if type(n[0]).__name__ == "nn":
            nt = n[0]
            namespace = nt.namespace
            name = nt.name

        elif n[0].data == "ma_default":
            default = n[0].children[0].data

        return self.nt_ma(namespace=namespace, name=name, default=default)

    def frag(self, n) -> tuple:
        """Transform frag tokens to nametuple."""
        frag_range, frag_descriptor = self._get_values(n, ("F_RANGE", "F_DESCRIPTOR"))
        nt = self.nt_frag(range=frag_range, descriptor=frag_descriptor)

        return self.nt_type("modifier", "fragment"), nt

    def loc(self, n) -> tuple:
        """Transform tokens to dict(function_name: namedtuple)."""
        return self.nt_type("modifier", "location"), n[0]

    def var(self, n) -> tuple:
        """Transform tokens to dict(function_name: namedtuple)."""
        return self.nt_type("modifier", "variant"), self.nt_var(n[0].value)

    def nn(self, n):
        """Transform tokens to dict(function_name: namedtuple)."""
        token_names = ("NAMESPACE_KEYWORD", ("NAME_WITHOUT_QUOTES", "STRING_SIMPLE"))
        namespace, name = self._get_values(n, token_names)

        token_dict = {x.type: x for x in n}
        name_with_quotes = token_dict.get("NAME_WITHOUT_QUOTES")
        simple_string = token_dict.get("STRING_SIMPLE")
        entry_token = name_with_quotes or simple_string

        self.cache.set_namespace_entry(namespace=namespace, entry=name, token=entry_token)

        return self.nt_nn(namespace=namespace, name=name)

    def pmod(self, n) -> dict:
        """Transform tokens to dict(function_name: namedtuple)."""
        namespace, name = "", ""

        pos_value = self._get_value(n, "POSITION")
        position = int(pos_value) if pos_value else 0

        nn = [x for x in n if x.__class__.__name__ == "nn"]
        if nn:
            namespace = nn[0].namespace
            name = nn[0].name

        aa = self._get_dict_value(n, "amino_acid")
        amino_acid = aa if aa else ""

        ptype = self._get_dict_value(n, "pmod_type")
        type_ = ptype if ptype else ""

        nt = self.nt_pmod(
            namespace=namespace,
            name=name,
            type=type_,
            amino_acid=amino_acid,
            position=position,
        )

        return self.nt_type("modifier", "pmod"), nt

    @staticmethod
    def amino_acid(n):
        """Return amino acid."""
        return {"amino_acid": n[0].data.split("_")[1].upper()}

    @staticmethod
    def pmod_type(n):
        """Return pmod_type."""
        return {"pmod_type": n[0].data}

    def from_loc(self, n):
        """Return tloc."""
        return [self.nt_type("translocation", "from_location"), n]

    def to_loc(self, n):
        """Return tloc."""
        return [self.nt_type("translocation", "to_location"), n]

    def rxn(self, n):
        """Return reaction."""
        return [self.nt_type("transformation", "reaction"), n]

    def reactants(self, n):
        """Return reactants."""
        return [self.nt_type("reaction_partner", "reactants"), sorted(n)]

    def products(self, n):
        """Return products."""
        return [self.nt_type("reaction_partner", "products"), sorted(n)]

    @staticmethod
    def fusion(n):
        """Return fusion."""
        return n[0]

    def _fusion_range(self, n):
        """Return fusion."""
        range_types = ("GENE_FUSION_RANGE", "RNA_FUSION_RANGE", "PROTEIN_FUSION_RANGE")
        fusion_range = self._get_value(n, range_types)
        return OrderedDict([("fusion_range", fusion_range)])

    def gene_fusion(self, n):
        """Return  gene fusion."""
        return [self.nt_type("other", "fusion_gene"), n]

    def gene_fusion_range(self, n):
        """Return fusion range."""
        return self._fusion_range(n)

    @staticmethod
    def fus_gene1(n):
        return ["gene1", n]

    @staticmethod
    def fus_gene2(n):
        return ["gene2", n]

    @staticmethod
    def fus_rna1(n):
        return ["rna1", n]

    @staticmethod
    def fus_rna2(n):
        return ["rna2", n]

    @staticmethod
    def fus_protein1(n):
        return ["protein1", n]

    @staticmethod
    def fus_protein2(n):
        return ["protein2", n]

    def rna_fusion(self, n):
        """Return rna fusion."""
        return [self.nt_type("other", "fusion_rna"), n]

    def rna_fusion_range(self, n):
        """Return fusion range."""
        return self._fusion_range(n)

    def protein_fusion(self, n):
        """Return protein fusion."""
        return [self.nt_type("other", "fusion_protein"), n]

    def protein_fusion_range(self, n):
        """Return fusion range."""
        return self._fusion_range(n)

    @staticmethod
    def _get_dict_value(n, key_name):
        """Return value of key_name for first dict with key_name."""
        for e in n:
            if isinstance(e, dict) and key_name in e:
                return e[key_name]

    @staticmethod
    def __get_token_dict(tokens: typing.List[Token]) -> dict:
        """Get dictionary of tokens with typs as key."""
        return {t.type: t.value for t in tokens if isinstance(t, Token)}

    @staticmethod
    def _get_value(tokens, token_name):
        """Get first Token value of list with token_name.

        If token_name is a list method checks aginst list.
        """
        token_names = (token_name,) if isinstance(token_name, str) else token_name
        for t in tokens:
            if isinstance(t, Token) and t.type in token_names:
                return t.value

    def _get_values(self, tokens, token_names: typing.Iterable) -> list:
        """Get list of token values in same order then token_names.

        If token name not exists empty str is returned.
        If token_names element is a Iterable of str, first token value is
        return from this list.
        """
        t_dict = self.__get_token_dict(tokens)
        ret_lst = []

        for t_name in token_names:
            if isinstance(t_name, str):
                ret_lst.append(t_dict.get(t_name, ""))

            elif isinstance(t_name, typing.Iterable):
                value = ""
                for name in t_name:
                    value = t_dict.get(name)
                    if value:
                        break
                ret_lst.append(value)

        return ret_lst

    @staticmethod
    def _get_tree(n: list, tree_name):
        """Return first lark.Tree, namedtuple or dict linked to tree_name."""
        for e in n:
            if isinstance(e, Tree) and e.data == tree_name:
                return e
            elif isinstance(e, dict) and len(e.keys()) == 1:
                return e[tree_name]

        raise "Type not supported or empty list"

    @staticmethod
    def _get_all_values_by_name(tree_or_tokens, token_name):
        n = tree_or_tokens
        if isinstance(tree_or_tokens, Tree):
            n = tree_or_tokens.children

        return sorted([x.value for x in n if isinstance(x, Token) and x.type == token_name])

    @staticmethod
    def _first_tree(n: list) -> Tree:
        """Return first tree in list."""
        for e in n:
            if isinstance(e, Tree):
                return e


class _BelScript:
    """Cache the content of the BEL script and methods to find errors and warnings."""

    def __init__(self, force_new_db):
        """Init."""
        # setup database
        engine = BelRdb().engine

        self.force_new_db = force_new_db
        reset_tables(engine, self.force_new_db)

        self._namespaces = Namespaces()  # entries Namespace objects
        self._annotations = Annotations()  # entries Annotation objects

        self.__namespace_in_db_updated = False
        self.__annotations_in_db_updated = False

        self._namespace_entries = NamespaceEntries()
        self._annotation_entries = AnnotationEntries()

        self.notDownloadedFromUrls = []

        self.namespace_manager = NamespaceManager(
            model=NamespaceModel,
            entries_model=NamespaceEntry,
            grammar_start=GRAMMAR_START_NS,
        )

        self.annotation_manager = AnnotationManager(
            model=AnnotationModel,
            entries_model=AnnotationEntry,
            grammar_start=GRAMMAR_START_ANNO,
        )

    def set_namespace_definition(self, as_type, keyword, value):
        """Set an annotation definition with type, keyword and value value could be 'file', 'url' or 'list'.

        :param str as_type: 'file', 'url' or 'list'
        :param str keyword: namespace keyword
        :param str value: URL, file path or list
        """
        if as_type in ALLOWED_TYPES:
            self._namespaces.add(as_type, keyword, value)
            return True
        else:
            logger.error("{} is not a allowed type of {}".format(as_type, ALLOWED_TYPES))
            return False

    def set_annotation_definition(self, as_type, keyword, value):
        """Set an annotation definition with type, keyword and value could be 'file', 'url' or 'list'.

        :param str as_type: 'file', 'url' or 'list'
        :param str keyword: namespace keyword
        :param str value: URL, file path or list
        """
        if as_type in ALLOWED_TYPES:
            self._annotations.add(as_type, keyword, value)
            return True
        else:
            logger.error("{} is not an allowed type of {}".format(as_type, ALLOWED_TYPES))
            return False

    def set_annotation_entry(self, annotation: str, entry: str, token: Token):
        """Set annotation, entry and lark.lexer.Token token.

        :param str annotation: annotation
        :param str entry: entry
        :param lark.lexer.Token token:
        """
        self._annotation_entries.set_annotation_entry(keyword=annotation, entry=entry, token=token)

    def set_namespace_entry(self, namespace: str, entry: str, token: Token):
        """Set namespace, entry and lark.lexer.Token token.

        :param str namespace:
        :param str entry:
        :param lark.lexer.Token token:
        """
        if not isinstance(token, Token):
            raise Exception("expecting Token in cache.set_namespace_entry")

        self._namespace_entries.set_namespace_entry(keyword=namespace, entry=entry, token=token)

    @property
    def errors(self) -> List[_Error]:
        """Execute all methods to find errors and warnings."""
        self.update_database()

        # all errors are children from errors._Error instances
        return (
            self.notDownloadedFromUrls
            + self.entries_without_namespace
            + self.entries_without_annotation
            + self.entries_not_in_namespace_url
            + self.entries_not_in_annotation_url
            + self.entries_not_in_namespace_list
            + self.entries_not_in_annotation_list
            + self.entries_not_in_namespace_pattern
            + self.entries_not_in_annotation_pattern
        )

    @property
    def warnings(self) -> List[_Warning]:
        """Execute all methods to find warnings."""
        if not (self.__namespace_in_db_updated and self.__annotations_in_db_updated):
            self.update_database()

        # all warnings are children from warnings._Warning instances
        return self.entries_also_in_other_namespace

    @property
    def entries_also_in_other_namespace(self) -> List[AlsoUsedInOtherNamespace]:
        """Return WithoutDefinedNamespace list."""
        ret = []
        # extract all entries used in BEL statements and create a dict of lower entries with all keywords
        entry_keyword_dict = defaultdict(set)
        for keyword1, entries in self._namespace_entries.entries.items():
            for entry in entries:
                entry_keyword_dict[entry.lower()] |= {keyword1}
        # identify all ambiguous entries (in more than 1 namespace)
        ambiguous_entries = {entry: keywords for entry, keywords in entry_keyword_dict.items() if len(keywords) > 1}

        # ToDo: iterate all lower entries an check for permutation
        #        for lower_entry in entry_keyword_dict:
        #            if lower_entry.count(",") == 1:
        #                reverse_without_comma = " ".join([x.strip() for x in lower_entry.split(",")][::-1])
        #                if reverse_without_comma in entry_keyword_dict:
        #                    print(lower_entry,
        #                    "%s exists in %s" % (reverse_without_comma,
        #                    entry_keyword_dict[reverse_without_comma]))
        #                    ret.append(AlsoUsedInOtherNamespace(keyword=keyword2,
        #                                                        entry=entry,
        #                                                        line_number=token.line,
        #                                                        column=token.column,
        #                                                        hint=hint))

        # iterate all tokens with namespace entries and check if they are also exists in ambiguous entries
        for keyword2, entries_tokens in self._namespace_entries.tokens.items():
            for entry, tokens in entries_tokens.items():
                if entry.lower() in ambiguous_entries:
                    ambiguous_tokens = self._namespace_entries.tokens[keyword2][entry]
                    for token in ambiguous_tokens:
                        hint = "%s exists also in %s" % (
                            entry,
                            ambiguous_entries[entry.lower()] - {keyword2},
                        )
                        ret.append(
                            AlsoUsedInOtherNamespace(
                                keyword=keyword2,
                                entry=entry,
                                line_number=token.line,
                                column=token.column,
                                hint=hint,
                            )
                        )
        return ret

    @property
    def entries_not_in_namespace_pattern(self) -> List[NotInNamespacePattern]:
        """Return a list of entries not fitting a given namespace pattern."""
        ret = []

        ns_pattern_kwds = self.used_namespace_keywords & self._namespaces.keywords_by_type(PATTERN)

        for kwd in ns_pattern_kwds:
            regex = self._namespaces.keyword_dict[kwd].value
            pattern = re.compile("^" + regex + "$")
            elcs = self._namespace_entries.get_entry_line_column_list_by_keyword(kwd)
            for entry, line, column in elcs:
                if not pattern.search(entry):
                    ret.append(NotInNamespacePattern(keyword=kwd, entry=entry, line_number=line, column=column))
        return ret

    @property
    def entries_not_in_annotation_pattern(self) -> List[NotInAnnotationPattern]:
        """Return a list of entries not fitting a given annotation pattern."""
        ret = []

        anno_pattern_kwds = self.used_annotation_keywords & self._annotations.keywords_by_type(PATTERN)

        for kwd in anno_pattern_kwds:
            regex = self._annotations.keyword_dict[kwd].value
            pattern = re.compile("^" + regex + "$")
            elcs = self._annotation_entries.get_entry_line_column_list_by_keyword(kwd)
            for entry, line, column in elcs:
                if not pattern.search(entry):
                    ret.append(NotInAnnotationPattern(keyword=kwd, entry=entry, line_number=line, column=column))
        return ret

    @property
    def entries_not_in_annotation_list(self) -> List[NotInAnnotationList]:
        """Return a list of entries not in a given annotations."""
        ret = []

        anno_kwd_used_and_as_list = self.used_annotation_keywords & self._annotations.keywords_by_type(LIST)

        for kwd in anno_kwd_used_and_as_list:
            elcs = self._annotation_entries.get_entry_line_column_list_by_keyword(kwd)
            for entry, line, column in elcs:
                if entry not in self._annotations.keyword_dict[kwd].value:
                    ret.append(NotInAnnotationList(keyword=kwd, entry=entry, line_number=line, column=column))
        return ret

    @property
    def entries_not_in_namespace_list(self) -> List[NotInNamespaceList]:
        """Return a list of entries not in a given namespace."""
        ret = []

        ns_kwd_used_and_as_list = self.used_namespace_keywords & self._namespaces.keywords_by_type(LIST)

        for kwd in ns_kwd_used_and_as_list:
            elcs = self._namespace_entries.get_entry_line_column_list_by_keyword(kwd)
            for entry, line, column in elcs:
                if entry not in self._namespaces.keyword_dict[kwd].value:
                    ret.append(NotInNamespaceList(keyword=kwd, entry=entry, line_number=line, column=column))
        return ret

    @property
    def entries_without_namespace(self) -> List[WithoutDefinedNamespace]:
        """Return WithoutDefinedNamespace list."""
        ret = []
        for namespace_keyword in self.namespaces_without_definition:
            elcs = self._namespace_entries.get_entry_line_column_list_by_keyword(namespace_keyword)
            for entry, line, column in elcs:
                ret.append(
                    WithoutDefinedNamespace(
                        keyword=namespace_keyword,
                        entry=entry,
                        line_number=line,
                        column=column,
                    )
                )
        return ret

    @property
    def entries_without_annotation(self) -> List[WithoutDefinedAnnotation]:
        """Return WithoutDefinedNamespace list."""
        ret = []
        for annotation_keyword in self.annotations_without_definition:
            elcs = self._annotation_entries.get_entry_line_column_list_by_keyword(annotation_keyword)
            for entry, line, column in elcs:
                ret.append(
                    WithoutDefinedAnnotation(
                        keyword=annotation_keyword,
                        entry=entry,
                        line_number=line,
                        column=column,
                    )
                )
        return ret

    def update_database(self) -> None:
        """Update namespace and annotation entries in database if not exists by url and keyword."""
        if not (self.__namespace_in_db_updated and self.__annotations_in_db_updated):
            self.__namespace_in_db_updated = self.update_namespaces_in_db()
            self.__annotations_in_db_updated = self.update_annotations_in_db()

    def set_entry_not_in_namespace_list_errors(self):
        pass

    def set_entry_not_in_annotation_list_errors(self):
        pass

    @property
    def entries_not_in_namespace_url(self) -> List[NotInNamespaceUrl]:
        """Return a list of entries not exists in namespaces referenced as URL.

        Returns
        -------
        List[NotInNamespaceUrl]
            Description of returned object.
        """
        entries_not_in_namespace = []

        for keyword in self.used_namespace_keywords:
            namespace = self._namespaces.keyword_dict[keyword]

            if namespace.as_type == URL:
                url = namespace.value
                elc_list = self._namespace_entries.get_entry_line_column_list_by_keyword(keyword)

                names_not_exists = self.namespace_manager.get_entries_not_exists(
                    keyword=keyword,
                    url=url,
                    entry_line_column_list=elc_list,
                )

                for entry, line, column, hint in names_not_exists:
                    error = NotInNamespaceUrl(
                        keyword=keyword,
                        url_or_path=url,
                        entry=entry,
                        line_number=line,
                        column=column,
                        hint=hint,
                    )
                    entries_not_in_namespace.append(error)

        return entries_not_in_namespace

    @property
    def entries_not_in_annotation_url(self) -> List[_Error]:
        """Return a list of entries not in the annotation URL."""
        entries_not_in_annotation = []

        for keyword in self.used_annotation_keywords:
            annotation = self._annotations.keyword_dict[keyword]

            if annotation.as_type == URL:
                url = annotation.value
                elc_list = self._annotation_entries.get_entry_line_column_list_by_keyword(keyword)

                names_not_exists = self.annotation_manager.get_entries_not_exists(
                    keyword=keyword, url=url, entry_line_column_list=elc_list
                )

                for entry, line, column, hint in names_not_exists:
                    error = NotInAnnotationUrl(
                        keyword=keyword,
                        url_or_path=url,
                        entry=entry,
                        line_number=line,
                        column=column,
                        hint=hint,
                    )
                    entries_not_in_annotation.append(error)
        return entries_not_in_annotation

    def update_annotations_in_db(self) -> bool:
        """Update annotation in database if URL and keyword not exists."""
        import_success = True
        for anno in self._annotations.to_update:
            if anno.keyword in self.used_annotation_keywords:
                if not self.annotation_manager.keyword_url_exists(keyword=anno.keyword, url=anno.value):
                    if anno.as_type == URL:
                        logger.info(f"Update db with annotation {anno.keyword}: download from {anno.value}")

                        (
                            successful,
                            error,
                        ) = self.annotation_manager.save_from_url_or_path(
                            keyword=anno.keyword,
                            url_or_path=anno.value,
                            doc_type=anno.as_type,
                        )

                        if not successful:
                            import_success = False
                            error_args = error.args[0].split("\n")
                            string_error = error_args[2] if len(error_args) > 1 else error_args[0]
                            logger.error(
                                f"Annotation {anno.keyword} failed to be added from {anno.value}",
                                exc_info=False,
                            )

                            if "column" in dir(error):  # Indicates it's a Lark error
                                download_error = NotDownloadedFromUrl(
                                    keyword=anno.keyword,
                                    url_or_path=anno.value,
                                    column=error.column,
                                    line=error.line,
                                    hint=f'{error.allowed} error in "{string_error}"',
                                )

                            else:  # It's an HTTPError of some kind
                                download_error = NotDownloadedFromUrl(
                                    keyword=anno.keyword,
                                    url_or_path=anno.value,
                                    column=0,
                                    line=0,
                                    hint=f"{string_error}",
                                )
                            self.notDownloadedFromUrls.append(download_error)

        return import_success

    def update_namespaces_in_db(self) -> bool:
        """Update namespaces in database if URL and keyword does not exist."""
        import_success = True
        for ns in self._namespaces.to_update:
            if ns.keyword in self.used_namespace_keywords:
                if not self.namespace_manager.keyword_url_exists(keyword=ns.keyword, url=ns.value):
                    if ns.as_type == URL:
                        logger.info(f"Update db with namespace {ns.keyword}: download from {ns.value}")

                        (
                            successful,
                            error,
                        ) = self.namespace_manager.save_from_url_or_path(
                            keyword=ns.keyword,
                            url_or_path=ns.value,
                            doc_type=ns.as_type,
                        )

                        if not successful:
                            import_success = False
                            error_args = error.args[0].split("\n")
                            string_error = error_args[2] if len(error_args) > 1 else error_args[0]
                            logger.error(
                                f"Namespace {ns.keyword} failed to be added from {ns.value}",
                                exc_info=False,
                            )

                            if "column" in dir(error):  # Indicates it's a Lark error
                                download_error = NotDownloadedFromUrl(
                                    keyword=ns.keyword,
                                    url_or_path=ns.value,
                                    column=error.column,
                                    line=error.line,
                                    hint=f'{error.allowed} error in "{string_error}"',
                                )

                            else:  # It's an HTTPError of some kind
                                download_error = NotDownloadedFromUrl(
                                    keyword=ns.keyword,
                                    url_or_path=ns.value,
                                    column=0,
                                    line=0,
                                    hint=f"{string_error}",
                                )

                            self.notDownloadedFromUrls.append(download_error)

        return import_success

    @property
    def namespaces_with_multiple_definitions(self):
        """Return all Namespace objects with several definitions.

        This is returned as a dictionary (key:keyword, value: list of Namespace objects).
        """
        ret = defaultdict(list)
        multiple_keyword = [k for k, v in Counter(self._namespaces.keywords).items() if v > 1]
        for ns in self._namespaces:
            if ns.keyword in multiple_keyword:
                ret[ns.keyword].append(ns)
        return dict(ret)

    @property
    def annotations_with_multiple_definitions(self):
        """Return all Annotation objects with several definitions.

        This is returned as a dictionary (key:keyword, value: list of Annotation objects).
        """
        ret = defaultdict(list)
        multiple_keyword = [k for k, v in Counter(self._annotations.keywords).items() if v > 1]
        for anno in self._annotations:
            if anno.keyword in multiple_keyword:
                ret[anno.keyword].append(anno)
        return dict(ret)

    @property
    def namespaces_without_definition(self):
        """Return set of namespace keywords used in statements but not defined with a reference.

        :return set: set of str
        """
        return set(self._namespace_entries.keywords) - set(self._namespaces.keywords)

    @property
    def annotations_without_definition(self):
        """Return a set of annotation keywords not defined with a reference.

        :return set: set of str
        """
        return set(self._annotation_entries.keywords) - set(self._annotations.keywords)

    @property
    def used_namespace_keywords(self) -> Set[str]:
        """Return set of used namespace keywords (with reference and used in statements)."""
        return set(self._namespace_entries.keywords) & set(self._namespaces.keywords)

    @property
    def used_annotation_keywords(self) -> Set[str]:
        """Return set of used namespace keywords."""
        return set(self._annotation_entries.keywords) & set(self._annotations.keywords)

    @property
    def namespace_keywords_in_statements(self):
        """Return all unique namespace keywords used in statements."""
        return self._namespace_entries.keywords

    @property
    def annotation_keywords_in_statements(self):
        """Return all unique annotation keywords used in statements."""
        return self._namespace_entries.keywords

    def get_entries_by_namespace_keyword(self, keyword):
        """Get all entries by namespace keyword.

        :param keyword: namespace keyword
        :return set: all entries in the namespace
        """
        return self._namespace_entries.get_entries_by_keyword(keyword)

    def get_entries_by_annotation_keyword(self, keyword):
        """Get all entries by namespace keyword.

        :param keyword: namespace keyword
        :return set: all entries in the namespace
        """
        return self._annotation_entries.get_entries_by_keyword(keyword)


class Entries:
    """Abstract class representing namespaces and annotations."""

    tokens = defaultdict(dict)
    entries = defaultdict(set)

    def get_entry_line_column_list_by_keyword(self, keyword: str) -> Generator[str, int, int]:
        """Get generator of tuple(entry, line, column) by keyword.

        Parameters
        ----------
        keyword: str
            Description of parameter `keyword: str`.

        Returns
        -------
        Generator
            Generator of tuple(entry: str, line: int, column: int).
        """
        for entry, tokens in self.tokens[keyword].items():
            for token in tokens:
                yield entry, token.line, token.column

    @property
    def keywords(self):
        """Return a list of unique keywords used in SETs."""
        return self.entries.keys()

    def get_entries_by_keyword(self, keyword: str) -> Set:
        """Get entries by keyword.

        :param str keyword: keyword to retrieve from dict
        """
        return self.entries.get(keyword, set())

    def get_tokens_by_keyword(self, keyword: str) -> Dict:
        """Get tokens by keyword.

        :param str keyword: keyword to retrieve from dict
        """
        return self.entries.get(keyword, set())

    def __str__(self):
        """String representation of object."""
        return str(dict(self.tokens))


class NamespaceEntries(Entries):
    """Namespace subclass of Entries."""

    def __init__(self):
        """Init."""
        self.entries = defaultdict(set)
        self.tokens = defaultdict(dict)

    def set_namespace_entry(self, keyword, entry, token):
        """Set namespace,  entry and lark.lexer.Token.

        :param str keyword: namespace
        :param str entry: entry
        :param lark.lexer.Token token: Token object from lark library
        """
        if isinstance(token, Token):
            self.entries[keyword] |= {entry}
            if keyword in self.tokens and entry in self.tokens[keyword]:
                self.tokens[keyword][entry].append(token)
            else:
                self.tokens[keyword][entry] = [token]
        else:
            raise "Argument token is type {} not {}".format(type(token), "lark.lexer.Token")


class AnnotationEntries(Entries):
    """Annotation subclass of Entries."""

    def __init__(self):
        """Init."""
        self.entries = defaultdict(set)
        self.tokens = defaultdict(dict)

    def set_annotation_entry(self, keyword: str, entry, token):
        """Set annotation,  entry and lark.lexer.Token.

        :param str keyword: annotation
        :param entry: entry
        :param lark.lexer.Token token: Token object from lark library
        """
        if isinstance(token, Token):
            self.entries[keyword] |= {entry}
            if keyword in self.tokens and entry in self.tokens[keyword]:
                self.tokens[keyword][entry].append(token)
            else:
                self.tokens[keyword][entry] = [token]
        else:
            raise "argument token is type {} not {}".format(type(token), "lark.lexer.Token")


class NsAnsBase:
    """Parent class for class Namespace and Annotation."""

    def __init__(self, obj_class):
        """Init."""
        self.__objs = []
        self.class_ = obj_class

    def add(self, as_type: str, keyword: str, value: str):
        """Add obj to list of objs.

        :param str as_type: allowed keywords 'file', 'url' or 'list'
        :param str keyword: keyword used in object
        :param str value: value of object
        :return:
        """
        obj = self.class_(as_type, keyword, value)
        self.__objs.append(obj)

    @property
    def type_dict(self) -> DefaultDict:
        """Convert to list of dictionaries."""
        ret = defaultdict(list)
        [ret[obj.as_type].append(obj) for obj in self]
        return ret

    def by_type(self, as_type: str):
        """Return list of Namespace objects by 'list', 'url' or 'file'."""
        if as_type not in ALLOWED_TYPES:
            raise "{} not in allowed types {}".format(as_type, ALLOWED_TYPES)
        return [obj for obj in self if obj.as_type == as_type]

    def keywords_by_type(self, as_type: str) -> Set[str]:
        """Return a set of keywords by Namespace type 'list', 'url' or 'file'."""
        if as_type not in ALLOWED_TYPES:
            raise "{} not in allowed types {}".format(as_type, ALLOWED_TYPES)
        return set([obj.keyword for obj in self if obj.as_type == as_type])

    @property
    def keyword_dict(self) -> Dict:
        """Return a dictionary of key=keyword, value: Namespace or Annotation object."""
        ret = dict()
        for obj in self:
            ret[obj.keyword] = obj
        return ret

    @property
    def keywords(self) -> List[str]:
        """Return all keywords."""
        return [obj.keyword for obj in self.__objs]

    @property
    def to_update(self) -> List:
        """Return a list of all Namespace or Annotation (NS_or_Anno) objects with URL or file path.

        :return list: list of all Namespace or Annotation (NS_or_Anno) objects with URL or file path
        """
        return self.type_dict[URL]

    def __iter__(self):
        """Return a generator of objects (Namespace or Annotation)."""
        for obj in self.__objs:
            yield obj


class Namespaces(NsAnsBase):
    """Namespace child class."""

    def __init__(self):
        """init."""
        super(Namespaces, self).__init__(obj_class=Namespace)


class Annotations(NsAnsBase):
    """Annotation child class."""

    def __init__(self):
        """init."""
        super(Annotations, self).__init__(obj_class=Annotation)


class Namespace:
    """Namespace class to represent BEL statement namespaces."""

    def __init__(self, as_type, keyword, value):
        """Namespace init."""
        self.as_type = as_type
        self.keyword = keyword
        self.value = value

    def to_dict(self):
        """Convert class values to dictionary."""
        return {"as_type": self.as_type, "keyword": self.keyword, "value": self.value}

    def __unicode__(self):
        return "Namespace:" + str(self.to_dict())

    def __str__(self):
        return self.__unicode__()


class Annotation:
    """Annotation class to represent BEL statement annotations."""

    def __init__(self, as_type, keyword, value):
        """Annotation init."""
        self.as_type = as_type
        self.keyword = keyword
        self.value = value

    def to_dict(self):
        """Convert class values to dictionary."""
        return {"as_type": self.as_type, "keyword": self.keyword, "value": self.value}

    def __unicode__(self):
        return "Annotation" + str(self.to_dict())

    def __str__(self):
        return self.__unicode__()



"""Methods for importing BEL files into OrientDB."""

import json
import logging
import os
import re
from collections import OrderedDict, defaultdict, namedtuple
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Tuple

import git
from git.exc import InvalidGitRepositoryError
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel import tools
from ebel.constants import RID
from ebel.manager.orientdb.constants import EDGES, NODES

from .odb_defaults import bel_func_short, normalized_pmod

BEL_GIT_ID = namedtuple("BEL_GIT_ID", ["hexsha", "repo_path", "origin_url"])
JsonParts = namedtuple("JsonParts", ["document", "definitions", "statements_and_sets"])

logger = logging.getLogger(__name__)


class _BelImporter:
    """BEL importer for OrientDB."""

    _cache = {}

    def __init__(self, file_path, odb_client: OrientDB = None):
        """Insert statements and sets from BEL JSON file."""
        self.client = odb_client
        self.file_path = file_path
        if not _BelImporter._cache:
            _BelImporter._cache[EDGES] = self._get_bel_relation_rid_cache()
            _BelImporter._cache[NODES] = self._get_bel_rid_cache()

    @staticmethod
    def file_is_in_git_repo(path):
        try:
            git.Repo(path, search_parent_directories=True)
            exists = True
        except InvalidGitRepositoryError:
            logger.error(f"{path} is not a valid Git repository")
            exists = False
        return exists

    def _get_bel_rid_cache(self):
        """Get BEL string OrientDB rid dictionary for all bel nodes."""
        sql = "Select @rid.asString(),@class.asString(),bel from bel"
        return {(y["bel"], y["class"]): y["rid"] for y in [x.oRecordData for x in self.client.command(sql)]}

    def _get_bel_relation_rid_cache(self):
        """Create a dictionary of all bel_relation edges using their properties as keys and RIDs as values."""
        rel_cache = dict()

        sql = """SELECT @rid.asString(), out.@rid.asString() as subject_id, in.@rid.asString() as object_id,
        evidence, annotation, citation.type as citation_type, citation.ref as citation_ref,
        @class.asString() as relation FROM bel_relation"""
        props = (
            "relation",
            "subject_id",
            "object_id",
            "citation_type",
            "citation_ref",
            "evidence",
            "annotation",
        )
        results = self.client.command(sql)
        for entry in results:
            r = entry.oRecordData
            edge_profile = []
            for prop in props:
                if prop in r:
                    if prop == "annotation":
                        edge_profile.append(json.dumps(r[prop], sort_keys=True))
                    else:
                        edge_profile.append(r[prop])

                else:
                    edge_profile.append(None)

            rel_cache[tuple(edge_profile)] = r["rid"]

        return rel_cache

    def get_git_info(self) -> BEL_GIT_ID:
        """Get git info identifies the file.

        Returns
        -------
        type
            object with properties hexsha, repo_path and origin_url

        """
        bel_git_id = None
        absolute_path = Path(self.file_path).absolute()

        if self.file_is_in_git_repo(absolute_path):
            repo = git.Repo(absolute_path, search_parent_directories=True)
            origin_url_found = re.search("(https?://)([^@]+@)?(.*)", repo.remotes.origin.url)

            origin_url = None
            if origin_url_found:
                protocol, user, origin_url = origin_url_found.groups()

            repo_path = absolute_path.as_posix()[len(repo.working_dir) + 1 :]
            try:
                commits = list(repo.iter_commits(paths=str(absolute_path)))

                if commits:
                    commit = commits[0]
                    bel_git_id = BEL_GIT_ID(commit.hexsha, repo_path, origin_url)
            except:
                print("Not able to get commit info")

        if not bel_git_id:
            bel_git_id = BEL_GIT_ID("", "", "")

        return bel_git_id

    def insert(self) -> Tuple[bool, int]:
        """Insert JSON file in OrientDB."""
        inserted = 0

        with open(self.file_path) as fd:
            bel_python_object = json.load(fd, object_pairs_hook=OrderedDict)

        if not bel_python_object:  # May be empty JSON
            logger.warning(f"{self.file_path} is empty")
            return False, 0

        parts = self.get_json_parts(bel_python_object)
        exists_before, document_id = self.insert_bel_header(parts.document, parts.definitions)

        if not exists_before:
            inserted = self.insert_statements_and_sets(parts.statements_and_sets, document_id)

        return not exists_before, inserted

    def insert_bel_header(self, doc_info, defs) -> Tuple[bool, str]:
        """Insert header info (document_info, namespaces and annotations) plus git info)."""
        data = deepcopy(doc_info)
        data["annotation"] = {}
        data["namespace"] = {}
        for entry in defs:
            anno_or_ns = list(entry.keys())[0]
            keyword = entry[anno_or_ns]["keyword"]
            del entry[anno_or_ns]["keyword"]
            data[anno_or_ns][keyword] = entry[anno_or_ns]

        if "authors" in data:
            data["authors"] = [a.strip() for a in data["authors"].split(",")]

        keyword_flag = False
        if "keywords" in data:
            keyword_flag = True
            data["keywords"] = [k.strip() for k in data["keywords"].split(",")]

            keyword_rids = self.get_keyword_rids(keywords=data["keywords"])
            keyword_linkset = "[" + ",".join(keyword_rids) + "]"  # This feature is a LINKSET
            del data["keywords"]  # TODO get so rids can be in JSON and load - see line 131

        now = datetime.today().strftime("%Y-%m-%d %H:%M:%S")
        data["date"] = {"uploaded": now}
        file_stat = os.stat(self.file_path)
        data["file"] = {
            "path": self.file_path,
            "md5": tools.md5(self.file_path),
            "size": file_stat.st_size,
            "last_modified": file_stat.st_mtime,
        }

        git_info_dict = self.get_git_info()

        data["git_info"] = git_info_dict._asdict()

        net_exists_before = self.network_exists(
            data["version"],
            self.file_path,
            data["file"]["md5"],
            data["file"]["last_modified"],
            git_info_dict,
        )

        if not net_exists_before:
            content = json.dumps(data)
            sql = "insert into bel_document content {}".format(content)
            r = self.client.command(sql)
            document_rid = r[0]._OrientRecord__rid

            # TODO How to include LINKSET in data dict? Fails when trying to load. This is work around
            if keyword_flag:
                keyword_sql = f"UPDATE {document_rid} SET keywords = {keyword_linkset}"
                self.client.command(keyword_sql)
        else:
            document_rid = None
            # TODO: Extract information from database
            logger.warning(f"{self.file_path} was already imported!")

        return net_exists_before, document_rid

    def get_keyword_rids(self, keywords: list) -> list:
        """Get the RID for each keyword set in the BEL document. If keyword does not exist, a new one is made."""
        rid_sql = "SELECT @rid FROM keyword WHERE label = '{}'"

        keyword_rids = []
        for keyword in keywords:
            record = self.client.command(rid_sql.format(keyword.lower()))

            if record:
                rid = record[0].oRecordData[RID].get_hash()

            else:
                insert_sql = f"INSERT INTO keyword CONTENT {{'label': '{keyword.lower()}', 'description': ''}}"
                new_record = self.client.command(insert_sql)[0]
                rid = new_record._OrientRecord__rid

            keyword_rids.append(rid)

        return keyword_rids

    def network_exists(self, version, file_path, md5, last_modified, git_info_dict) -> bool:
        sql = f"""Select 1 from bel_document where
                    version = '{version}' and
                    file.path = '{Path(file_path).as_posix()}' and
                    file.md5 = '{md5}' and
                    file.last_modified = '{last_modified}' and
                    git_info = {json.dumps(git_info_dict)}
                    """
        return not len(self.client.command(sql)) == 0

    @staticmethod
    def insert_definitions(definitions):
        # TODO why is this here
        return definitions

    def insert_statements_and_sets(self, statements_and_sets, document_id) -> int:
        """Insert statement and sets."""
        citation = {}
        evidence = ""
        annotation = defaultdict(set)
        pmid = 0
        citation_ref = ""
        citation_type = ""

        for e in tqdm(statements_and_sets, desc="Insert BEL Statements"):
            dtype, data = tuple(e.items())[0]

            if dtype == "sets":
                for bel_set in data:
                    key, value = tuple(bel_set.items())[0]

                    if key == "citation":
                        citation = dict(value)
                        citation_type = citation["type"].strip()
                        citation_ref = citation["ref"].strip()
                        evidence = ""
                        annotation = defaultdict(set)

                        if citation["type"].lower() == "pubmed" and re.search(r"^\d+$", citation_ref):
                            pmid = citation_ref
                        else:
                            pmid = 0

                    elif key == "evidence":
                        evidence = re.sub(r"\s*\\\s*\n\s*", " ", value)

                    elif key == "set":
                        anno_keyword, anno_entries = tuple(value.items())[0]
                        annotation[anno_keyword] = set(anno_entries)

                    elif key == "unset":
                        for anno_keyword in value:
                            annotation.pop(anno_keyword, None)

            elif dtype == "statement" and len(data) >= 1:
                subject_id = self.get_node_id(data[0]["subject"])[1]

                if len(data) > 1 and "object" in data[2]:
                    # TODO: nested statements are missing

                    object_id = self.get_node_id(data[2]["object"])[1]

                    relation = data[1]["relation"]

                    self.insert_bel_edge(
                        annotation,
                        citation,
                        citation_ref,
                        citation_type,
                        evidence,
                        object_id,
                        pmid,
                        relation,
                        subject_id,
                        document_id,
                    )

        return len(statements_and_sets)

    def insert_bel_edge(
        self,
        annotation,
        citation,
        citation_ref,
        citation_type,
        evidence,
        object_id,
        pmid,
        relation,
        subject_id,
        document_id,
    ):
        # check if relation already exists
        anno = {key: sorted(list(annotation[key])) for key in annotation.keys()}
        anno_json = json.dumps(anno, sort_keys=True)
        edge_content = {
            "pmid": pmid,
            "citation": citation,
            "annotation": anno,
            "evidence": evidence,
            "document": [document_id],
        }

        # Need to clean the properties
        evidence = evidence.replace("\n", " ")
        citation_ref = citation_ref if citation_ref else None
        citation_type = citation_type if citation_type else None

        edge_profile = (
            relation,
            subject_id,
            object_id,
            citation_type,
            citation_ref,
            evidence,
            anno_json,
        )
        edge_exists = True if edge_profile in _BelImporter._cache[EDGES] else False

        if not edge_exists:
            edge_content_json = json.dumps(edge_content, sort_keys=True)
            sql = f"CREATE EDGE {relation} FROM {subject_id} TO {object_id} CONTENT {edge_content_json}"
            record = self.client.command(sql)
            _BelImporter._cache[EDGES][edge_profile] = record[0]._OrientRecord__rid

        else:
            edge_rid = _BelImporter._cache[EDGES][edge_profile]
            sql = f"UPDATE EDGE {edge_rid} ADD document = {document_id}"
            self.client.command(sql)

    @classmethod
    def get_json_parts(cls, bel_python_object) -> JsonParts:
        """Return the parts document,definitions,statements_and_sets from bel_json as python object."""
        doc, defs, ss = bel_python_object
        return JsonParts(
            document=doc["document"],
            definitions=defs["definitions"],
            statements_and_sets=ss["statements_and_sets"],
        )

    @staticmethod
    def get_node_class_rid_from_db(obj, bel: str):
        """Return @rid if node exists."""
        rid = None
        node_class = None

        if isinstance(obj[0], OrderedDict) and "function" in obj[0]:
            node_class = obj[0]["function"]["name"]
            if (bel, node_class) in _BelImporter._cache[NODES]:
                rid = _BelImporter._cache[NODES][(bel, node_class)]

        return node_class, rid

    @staticmethod
    def is_function(obj):
        return isinstance(obj, OrderedDict) and "function" in obj

    def get_node_id(self, obj: list) -> str:
        """Return rid of obj."""
        if not isinstance(obj, list):
            raise TypeError(f"Expecting list, but get {type(obj)} for {obj}")

        inserted_nodes = []
        params = {}

        bel = self.get_bel(obj)

        node_class = obj[0]["function"]["name"]

        if node_class not in ["pmod", "fragment", "variant"]:
            node_class, rid_from_db = self.get_node_class_rid_from_db(obj, bel)
            if rid_from_db:
                return node_class, rid_from_db

        [params.update(x) for x in obj[1] if isinstance(x, OrderedDict)]

        for e in obj[1]:
            if isinstance(e, list):
                if self.is_function(e[0]):
                    inserted_nodes += [self.get_node_id(e)]
                else:
                    for f in e:
                        inserted_nodes += [self.get_node_id(f)]

        rid = self.insert_bel_node(node_class, params, bel)

        for child_class, child_rid in inserted_nodes:
            sql = f"Select @rid from has__{child_class} where in.@rid={child_rid} and out.@rid={rid}"
            exists = self.client.command(sql)
            if not exists:
                sql = f"CREATE EDGE has__{child_class} from {rid} to {child_rid}"
                self.client.command(sql)

        return node_class, rid

    def insert_bel_node(self, node_class, params, bel):
        """Insert bel node, return rid."""
        params.update({"bel": bel})
        sql_temp = "insert into {} content {}"
        sql = sql_temp.format(node_class, json.dumps(params))
        record = self.client.command(sql)
        rid = record[0]._OrientRecord__rid
        _BelImporter._cache[NODES][(bel, node_class)] = rid
        return rid

    def get_bel_string(self, params, function_name=None):
        """Get BEL formated string."""
        bels = []

        for param in params:
            if isinstance(param, str):
                bels.append(param)
            elif isinstance(param, dict):
                if set(param.keys()) == {"namespace", "name"}:
                    bels.append(param["namespace"] + ':"' + param["name"] + '"')

                elif function_name == "fragment":
                    bels.append(",".join(['"' + x + '"' for x in param.values() if x]))

                elif function_name == "activity":
                    if param["namespace"]:
                        bel_str = param["namespace"] + ':"' + param["name"] + '"'
                    else:
                        bel_str = param["default"]
                    bels.append("ma(" + bel_str + ")")

                elif function_name == "pmod":
                    if param["namespace"]:
                        first_part_pmod = param["namespace"] + ':"' + param["name"] + '"'
                    else:
                        first_part_pmod = normalized_pmod[param["type"]]
                    position = str(param["position"]) if param["position"] else None
                    parts_pmod = [first_part_pmod, param["amino_acid"], position]
                    bels.append(",".join([x for x in parts_pmod if x]))

                else:
                    bels.append(",".join(['"' + str(x) + '"' for x in param.values() if x]))

        joined_params = ",".join(bels)

        if function_name:
            return bel_func_short[function_name] + "(" + joined_params + ")"

        else:
            return joined_params

    def get_bel(self, obj, parent_function=None):
        """Return BEL by python object loaded from JSON."""
        params = []
        function_name = None

        for element in obj:
            if isinstance(element, dict):
                if "function" in element:
                    function_name = element["function"]["name"]

                else:
                    params.append(element)

            elif isinstance(element, list):
                params.append(self.get_bel(element, function_name))

        return self.get_bel_string(params, parent_function)



"""Reactome API methods."""

from flask import request

from ebel.manager.rdbms.models import reactome
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_reactome():
    """Get generic Reactome entry."""
    return _get_data(reactome.Reactome)


def get_bel_node_by_pathway_name():
    """Get Reactome related eBEL nodes by pathway name."""
    pathway_name = request.args.get("pathway_name")
    sql = f"""SELECT
            @rid.asString() as rid,
            namespace,
            name,
            bel,
            reactome_pathways
        FROM
            protein
        WHERE
            pure=true AND
            "{pathway_name}" in reactome_pathways
    """
    return _get_paginated_ebel_query_result(sql)



"""IntAct RDBMS model definition."""
from sqlalchemy import Column, Float, Integer, String, Text
from sqlalchemy.ext.declarative import declarative_base

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class Intact(Base):
    """Class definition for the intact table."""

    __tablename__ = "intact"
    id = Column(Integer, primary_key=True)
    confidence_value = Column(Float, index=True)
    detection_method = Column(String(100), index=True)
    detection_method_psimi_id = Column(Integer)
    int_a_uniprot_id = Column(String(50), index=True)
    int_b_uniprot_id = Column(String(50), index=True)
    interaction_ids = Column(Text)
    interaction_type = Column(String(100), index=True)
    interaction_type_psimi_id = Column(Integer)
    pmid = Column(Integer)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)



"""Bel module tests."""

from .constants import TEST_JSON, test_client
from .true_values import EDGES_EXTENSION, NODES_EXTENSION

bel = test_client


class TestBel:
    """Test class for bel module."""

    # def test_create_ssfs(self):
    #     """Test that server side functions are properly made."""
    #     bel.clear_server_side_functions()
    #     assert not bel.server_side_function_names()  # Should be no SSFs
    #     bel.create_ssfs()
    #     ssf_names = bel.server_side_function_names()
    #     # Check that the number of functions in db is equal or greater than in odb_ss_functions.py
    #     assert len(ssf_names) == len(ssfs.keys())
    #
    #     # Check that all in odb_ss_functions.py are in db
    #     base_ssfs = [func['name'] for func in ssfs.values()]
    #     assert all([func in base_ssfs for func in ssf_names])

    def test_bel_properties(self):
        """Test that all properties are present."""
        assert hasattr(bel, "hgnc")
        assert hasattr(bel, "uniprot")
        # assert hasattr(bel, 'dea')
        assert hasattr(bel, "drugbank")
        assert hasattr(bel, "gwas_catalog")
        assert hasattr(bel, "reactome")
        assert hasattr(bel, "biogrid")
        assert hasattr(bel, "stringdb")
        assert hasattr(bel, "clinical_trials")
        assert hasattr(bel, "intact")
        assert hasattr(bel, "clinvar")
        assert hasattr(bel, "mirtarbase")
        assert hasattr(bel, "disgenet")
        assert hasattr(bel, "pathway_commons")
        assert hasattr(bel, "kegg")
        assert hasattr(bel, "ensembl")
        assert hasattr(bel, "iuphar")
        assert hasattr(bel, "chebi")
        assert hasattr(bel, "nsides")
        assert hasattr(bel, "ncbi")
        assert hasattr(bel, "protein_atlas")

    def test_import_json(self):
        """Tests update_from_protein2gene feature of import_json."""
        # Clear all nodes and edges from test db
        bel.clear_all_nodes_and_edges()
        bel.clear_documents()

        assert all([count == 0 for count in bel.number_of_nodes.values()])
        assert all([count == 0 for count in bel.number_of_edges.values()])

        files_imported = bel.import_json(input_path=TEST_JSON, update_from_protein2gene=True, extend_graph=False)

        assert len(files_imported) == 1
        assert bel.number_of_nodes == NODES_EXTENSION
        assert bel.number_of_edges == EDGES_EXTENSION



"""DrugBank."""

import getpass
import logging
import os
import os.path
import platform
import re
import signal
from collections import namedtuple
from configparser import ConfigParser
from datetime import datetime
from time import time
from typing import Dict, Optional
from zipfile import ZipFile

import requests
from lxml.etree import iterparse
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.config import get_config_as_dict, write_to_config
from ebel.constants import DATA_DIR, RID
from ebel.defaults import config_file_path
from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.constants import DRUGBANK
from ebel.manager.rdbms.models import drugbank

ACTIONS = "actions"
SYMBOLS = "symbols"
KNOWN_ACTION = "known_action"

logger = logging.getLogger(__name__)

XML_URL = "http://www.drugbank.ca"
XML_NAMESPACE = f"{{{XML_URL}}}"
XN = {"n": XML_URL}

Xpath = namedtuple(
    "Xpath",
    (
        "references",
        "synonyms",
        "product_names",
        "pathways",
        "target_known_action",
        "ex_ids_resourec",
        "ex_ids_id",
        "target_actions",
        "target_uniprot",
    ),
)


class DrugBank(odb_meta.Graph):
    """Drugbank."""

    def __init__(self, client: Optional[OrientDB] = None):
        """Drugbank database.

        documentation: https://www.drugbank.ca/documentation
        """
        self.client = client
        self.biodb_name = DRUGBANK
        self.urls = {
            "drugbank_version": urls.DRUGBANK_VERSION,
            self.biodb_name: urls.DRUGBANK_DATA,
        }
        super().__init__(
            nodes=odb_structure.drugbank_nodes,
            edges=odb_structure.drugbank_edges,
            tables_base=drugbank.Base,
            indices=odb_structure.drugbank_indices,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

        self.biodb_dir = os.path.join(DATA_DIR, self.biodb_name)
        os.makedirs(self.biodb_dir, exist_ok=True)

        self.file_path = os.path.join(self.biodb_dir, "drugbank_all_full_database.xml.zip")
        self.file_name_unzipped = "full database.xml"
        self.file_path_unzipped = os.path.join(self.biodb_dir, self.file_name_unzipped)

        self.xpath_pattern = Xpath(
            references="./n:articles/n:article/n:pubmed-id/text()",
            synonyms="./n:synonym/text()",
            product_names="./n:product/n:name/text()",
            ex_ids_resourec="./n:resource[1]/text()",
            ex_ids_id="./n:identifier[1]/text()",
            target_actions="./n:actions[1]/n:action[1]/text()",
            target_known_action="./n:known-action[1]/text()",
            target_uniprot="./n:polypeptide[@source='Swiss-Prot']/@id",
            pathways="./n:pathway/n:smpdb-id/text()",
        )

    def __len__(self):
        """Get number of 'has_drug_target' graph edges."""
        return self.execute("Select count(*) from has_drug_target")[0].oRecordData["count"]

    def __contains__(self, unique_identifier):
        """Check if drug with identifier '?' exists in graph."""
        pass

    def _unzip_file(self):
        logger.info(f"unzip {self.file_path}")
        with ZipFile(self.file_path, "r") as zip_ref:
            zip_ref.extract(self.file_name_unzipped, self.biodb_dir)

    def insert_data(self) -> Dict[str, int]:
        """Insert data and returns number of inserts."""
        logger.info("Insert DrugBank data")

        self._unzip_file()

        drug_index = 0
        columns = [
            "name",
            "description",
            "cas-number",
            "unii",
            "state",
            "indication",
            "pharmacodynamics",
            "toxicity",
            "metabolism",
            "absorption",
            "half-life",
            "route-of-elimination",
            "volume-of-distribution",
            "clearance",
            "mechanism-of-action",
            "fda-label",
        ]

        doc = iterparse(self.file_path_unzipped, events=("end",), tag=f"{XML_NAMESPACE}drug")
        for action, elem in tqdm(doc, desc=f"Import {self.biodb_name.upper()}"):
            if "type" in elem.attrib:
                drug_index += 1
                drug = {"id": drug_index}
                references, synonyms = None, None
                targets = []
                external_identifiers = []
                product_names = []
                drug_interactions = []
                statuses = []
                patents = []
                pathways = []

                for child in elem.iterchildren():
                    ctag = child.tag[len(XML_NAMESPACE) :]

                    if ctag in columns:
                        drug[ctag.replace("-", "_")] = child.text

                    elif ctag == "drugbank-id":
                        primary_id = child.attrib.get("primary")
                        if primary_id:
                            drug["drugbank_id"] = child.text

                    elif ctag == "general-references":
                        pmid_strs = child.xpath(self.xpath_pattern.references, namespaces=XN)
                        references = [drugbank.Reference(pmid=int(x)) for x in pmid_strs]

                    elif ctag == "synonyms":
                        syns = child.xpath(self.xpath_pattern.synonyms, namespaces=XN)
                        synonyms = [drugbank.Synonym(synonym=x) for x in set(syns)]

                    elif ctag == "products":
                        pro_names = child.xpath(self.xpath_pattern.product_names, namespaces=XN)
                        product_names = [drugbank.ProductName(name=x) for x in set(pro_names)]

                    elif ctag == "drug-interactions":
                        for di_child in child:
                            di = {x.tag[len(XML_NAMESPACE) :].replace("-", "_"): x.text for x in di_child}
                            drug_interactions.append(drugbank.DrugInteraction(**di))

                    elif ctag == "external-identifiers":
                        for ex_ids_child in child:
                            resource = ex_ids_child.xpath(self.xpath_pattern.ex_ids_resourec, namespaces=XN)[0]
                            identifier = ex_ids_child.xpath(self.xpath_pattern.ex_ids_id, namespaces=XN)[0]
                            external_identifiers.append(
                                drugbank.ExternalIdentifier(resource=resource, identifier=identifier)
                            )

                    elif ctag == "targets":
                        for target in child:
                            u = target.xpath(self.xpath_pattern.target_uniprot, namespaces=XN)
                            uniprot = u[0] if u else None

                            if uniprot:
                                actions = target.xpath(self.xpath_pattern.target_actions, namespaces=XN)
                                action = actions[0] if actions else None

                                kactions = target.xpath(
                                    self.xpath_pattern.target_known_action,
                                    namespaces=XN,
                                )
                                known_action = kactions[0] if kactions else None

                                targets.append(
                                    drugbank.Target(
                                        uniprot=uniprot,
                                        action=action,
                                        known_action=known_action,
                                    )
                                )

                    elif ctag == "patents":
                        for patent in child:
                            patent_dict = {}
                            for patent_child in patent:
                                patent_key = patent_child.tag[len(XML_NAMESPACE) :].replace("-", "_")
                                patent_value = patent_child.text
                                if patent_key in ("expires", "approved") and re.search(
                                    r"^\d{4}-\d{2}-\d{2}$", patent_value
                                ):
                                    patent_value = datetime.strptime(patent_value.strip(), "%Y-%m-%d").date()
                                patent_dict[patent_key] = patent_value
                            if patent_dict:
                                patents.append(drugbank.Patent(**patent_dict))

                    elif ctag == "groups":
                        for group in child:
                            statuses.append(drugbank.Status(status=group.text))

                    elif ctag == "pathways":
                        pws = child.xpath(self.xpath_pattern.pathways, namespaces=XN)
                        pathways = [drugbank.Pathway(smpdb_id=x) for x in pws]

                    # TODO: implement reactions, ... (check xsd/xml)

                drugbank_instance = drugbank.Drugbank(
                    references=references,
                    synonyms=synonyms,
                    targets=targets,
                    external_identifiers=external_identifiers,
                    product_names=product_names,
                    drug_interactions=drug_interactions,
                    statuses=statuses,
                    patents=patents,
                    pathways=pathways,
                    **drug,
                )
                self.session.add(drugbank_instance)
                # insert in chunks of ...
                if drug_index % 100 == 0:
                    self.session.commit()

            elem.clear()

        self.session.commit()

        os.remove(self.file_path_unzipped)

        return {self.biodb_name: drug_index}

    @staticmethod
    def get_timed_answer(prompt=None, timeout=20, timeout_msg=None):
        """Returns user answer in an allotted amount of time."""

        def timeout_error(*_):
            raise TimeoutError

        if platform.system() == "Windows":
            answer = input(prompt)
            return answer

        else:
            signal.signal(signal.SIGALRM, timeout_error)
            signal.alarm(timeout)
            try:
                answer = input(prompt)
                signal.alarm(0)
                return answer

            except TimeoutError:
                if timeout_msg:
                    logger.error(f"The following error occurred while contacting the DrugBank website: {timeout_msg}")
                signal.signal(signal.SIGALRM, signal.SIG_IGN)
                return None

    def get_user_passwd(
        self,
        drugbank_user: Optional[str] = None,
        drugbank_password: Optional[str] = None,
    ) -> Optional[list]:
        """Read username and password from configuration file."""
        section_name = "DRUGBANK"
        conf = None
        section_exists = True
        config_exists = os.path.isfile(config_file_path)

        cfg = ConfigParser()

        if config_exists:
            cfg.read(config_file_path)

            if not cfg.has_section(section_name):
                section_exists = False

            else:
                conf = get_config_as_dict()[section_name]

        if not section_exists and (drugbank_user and drugbank_password):
            conf = {"user": drugbank_user, "password": drugbank_password}
            write_to_config(section_name, "user", drugbank_user)
            write_to_config(section_name, "password", drugbank_password)

        if not section_exists or not config_exists:
            prompt = "Do you have an approved account with DrugBank [y/n]: "
            timeout_msg = "No answer was provided. Skipping DrugBank update...\n"
            timeout = 20
            answer = self.get_timed_answer(prompt=prompt, timeout=timeout, timeout_msg=timeout_msg)

            num_tries = 0
            max_tries = 4

            while num_tries < max_tries and answer not in ["yes", "y", "no", "n"] and answer is not None:
                print("Invalid response!\n")
                answer = self.get_timed_answer(prompt=prompt, timeout=timeout, timeout_msg=timeout_msg)
                num_tries += 1

                if num_tries == max_tries:
                    print("Too many invalid responses. Skipping DrugBank update...\n")
                    num_tries += 1

            if answer in ["yes", "y"]:
                user = input("Insert DrugBank user name:\n")
                passwd = getpass.getpass("Insert DrugBank password:\n")
                write_to_config(section_name, "user", user)
                write_to_config(section_name, "password", passwd)
                conf = get_config_as_dict()[section_name]

            elif answer in ["no", "n"] or num_tries >= max_tries or answer is None:  # If no, write 'NA' to config
                write_to_config(section_name, "user", "NA")
                write_to_config(section_name, "password", "NA")

        if conf and conf["user"] != "NA":  # Return None if user in config is 'NA'
            return [conf["user"], conf["password"]]

    def insert(self) -> Dict[str, int]:
        """Check if files missing for download or generic table empty. If True then insert data."""
        inserted = {}
        self.download()

        if os.path.exists(self.file_path):
            drugbank_table_exists = self.engine.dialect.has_table(
                self.engine.connect(), drugbank.Drugbank.__tablename__
            )
            if not drugbank_table_exists or self.session.query(drugbank.Drugbank.id).count() == 0:
                self.recreate_tables()
                inserted.update(self.insert_data())
        else:
            logger.error(
                "Drugbank data can not be inserted because the file does not exist. Please download manually "
                f"and move to {self.biodb_dir}"
            )
        return inserted

    def download(self) -> dict:
        """Download DrugBank to standard data dir."""
        secs_100_days = 100 * 60 * 60 * 24
        now = time()
        downloaded = False
        file_exists = os.path.exists(self.file_path)

        if file_exists:
            expired = now - os.path.getmtime(self.file_path) > secs_100_days
        else:
            expired = True

        if not file_exists or expired:
            credentials = self.get_user_passwd()
            if credentials:
                file_to_download = self.__latest_version()
                logger.info(f"Downloading {file_to_download}")
                user, passwd = credentials[0], credentials[1]
                response = requests.get(file_to_download, auth=(user, passwd))

                if response.ok and response.status_code == 200:  # The file is not empty
                    with open(self.file_path, "wb") as drugbank_data_file:
                        drugbank_data_file.write(response.content)
                    downloaded = True

                else:
                    logger.warning("Invalid username and password. Skipping DrugBank update...")

        return {self.biodb_name: downloaded}

    def __latest_version(self) -> str:
        """Gets latest version of Drugbank and adds it to URL."""
        webpage = requests.get(self.urls["drugbank_version"])
        version = re.findall(r"(DrugBank Release Version) (\d\.\d\.\d)", webpage.text)[0][1]
        return self.urls[self.biodb_name].format(version.replace(".", "-"))

    def get_drugbank_id_rids(self) -> Dict[str, str]:
        """Get dict of DrugBank IDs as keys and their rIDs as values."""
        rows = self.execute("Select drugbank_id, @rid.asString() as rid from drug")
        return {x.oRecordData["drugbank_id"]: x.oRecordData[RID] for x in rows if "drugbank_id" in x.oRecordData}

    @staticmethod
    def _replace_new_lines(input_string: str) -> str:
        if input_string and input_string.strip():
            return input_string.strip().replace("\r\n", "<br>").replace("\n", "<br>")
        return input_string

    def update_interactions(self) -> int:
        """Updates the edges between drug and pure protein nodes."""
        self.clear_edges()

        drugbank_table_exists = self.engine.dialect.has_table(self.engine.connect(), drugbank.Drugbank.__tablename__)
        if not drugbank_table_exists:
            logger.error("Update failed - DrugBank table does not exist.")
            return 0

        drugbank_id_rids = self.get_drugbank_id_rids()

        inserted = 0

        sql = """Select @rid.asString() as rid, uniprot FROM protein
        WHERE pure=true and uniprot IS NOT NULL AND namespace = 'HGNC'"""
        for row in tqdm(self.execute(sql), desc=f"Update {self.biodb_name.upper()} interaction."):
            r = row.oRecordData
            protein_rid, uniprot = r["rid"], r["uniprot"]
            query = self.session.query(drugbank.Target).filter(drugbank.Target.uniprot == uniprot)
            for target in query.all():
                drugbank_id = target.drugbank.drugbank_id

                if drugbank_id in drugbank_id_rids:
                    drug_rid = drugbank_id_rids[drugbank_id]
                else:
                    value_dict_drug = {
                        "drugbank_id": target.drugbank.drugbank_id,
                        "label": target.drugbank.name,
                        "description": self._replace_new_lines(target.drugbank.description),
                        "cas_number": target.drugbank.cas_number,
                        "indication": self._replace_new_lines(target.drugbank.indication),
                        "pharmacodynamics": self._replace_new_lines(target.drugbank.pharmacodynamics),
                        "toxicity": self._replace_new_lines(target.drugbank.toxicity),
                        "metabolism": self._replace_new_lines(target.drugbank.metabolism),
                        "mechanism_of_action": self._replace_new_lines(target.drugbank.mechanism_of_action),
                    }

                    drug_rid = self.insert_record("drug_db", value_dict=value_dict_drug)
                    drugbank_id_rids[drugbank_id] = drug_rid  # update cache

                value_dict_edge = {
                    "action": target.action,
                    "known_action": target.known_action,
                }
                self.create_edge(
                    "has_drug_target_db",
                    from_rid=drug_rid,
                    to_rid=protein_rid,
                    value_dict=value_dict_edge,
                )
                inserted += 1

        return inserted



"""BioGrid."""

import typing
from enum import Enum
from typing import Dict, Tuple

import numpy as np
import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel import tools
from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.constants import BIOGRID
from ebel.manager.orientdb.odb_defaults import BelPmod, normalized_pmod_reverse
from ebel.manager.rdbms.models import biogrid

STANDARD_NAMESPACES = {9606: "HGNC", 10090: "MGI", 10116: "RGD"}


class BioGridNode:
    """Custom class definition for BioGRID nodes."""

    def __init__(
        self,
        rid: str,
        symbol: str,
        uniprot: str,
        taxonomy_id: int,
        pmod_bel: str = None,
    ):
        """Init for BioGRID generated nodes."""
        self.rid = rid
        self.symbol: str = symbol
        self.uniprot: str = uniprot
        self.taxonomy_id: int = taxonomy_id
        self.namespace, self.name = self.__get_nn()
        self.pmod_bel = pmod_bel

    def __get_nn(self) -> Tuple[str, str]:
        if self.taxonomy_id in STANDARD_NAMESPACES:
            namespace = STANDARD_NAMESPACES[self.taxonomy_id]
            name = self.symbol
        else:
            namespace = "UNIPROT"
            name = self.uniprot
        return namespace, name

    def __get_protein_as_value_dict(self, pmod: bool = False):
        if pmod:
            bel = self.get_pmod_protein_bel()
        else:
            bel = self.get_protein_bel()

        value_dict = {
            "name": self.name,
            "namespace": self.namespace,
            "bel": bel,
            "uniprot": self.uniprot,
        }
        return value_dict

    def get_pmod_protein_as_value_dict(self):
        """Get all pmod node metadata as dictionary."""
        return self.__get_protein_as_value_dict(pmod=True)

    def get_protein_as_value_dict(self):
        """Get all protein node metadata as dictionary."""
        return self.__get_protein_as_value_dict(pmod=False)

    def get_pmod_protein_bel(self) -> str:
        """Get the namespace and value for a protein and it's protein modification."""
        return f'p({self.namespace}:"{self.name}",pmod({self.pmod_bel}))'

    def get_protein_bel(self) -> str:
        """Get the namespace and value for a protein."""
        return f'p({self.namespace}:"{self.name}")'

    def get_pmod_bel(self):
        """Get the protein modification for a BEL."""
        return {
            "bel": f"pmod({self.pmod_bel})",
            "type": normalized_pmod_reverse[self.pmod_bel],
        }


class BioGridEdge:
    """Class definition for BioGRID edges."""

    def __init__(
        self,
        subject_rid: str,
        subject_symbol: str,
        subject_uniprot: str,
        subject_taxonomy_id: int,
        modification: str,
        object_rid: str,
        object_symbol: str,
        object_uniprot: str,
        object_taxonomy_id: int,
        experimental_system: str,
        pmids: str,
        num_pubs: int,
        dois: str,
        biogrid_ids: str,
    ):
        """Init for BioGRID edges."""
        mods = Modification.get_reverse_dict()
        self.modification = modification

        self.modConfig: ModConfig = mods[modification]

        self.subj: BioGridNode = BioGridNode(
            subject_rid,
            subject_symbol,
            subject_uniprot,
            subject_taxonomy_id,
            self.modConfig.pmod_bel,
        )
        self.obj: BioGridNode = BioGridNode(
            object_rid,
            object_symbol,
            object_uniprot,
            object_taxonomy_id,
            self.modConfig.pmod_bel,
        )

        self.experimental_system: str = experimental_system
        self.pmids: str = pmids
        self.num_pubs: int = num_pubs
        self.dois: str = dois
        self.biogrid_ids = biogrid_ids

    @property
    def both_pure_proteins_exists(self):
        """Check if both subject and object exist."""
        return bool(self.subj.rid) and bool(self.obj.rid)

    def get_edge_value_dict(self):
        """Get edge value dictionary."""
        value_dict = {"modification": self.modification}
        if self.dois:
            value_dict["dois"] = self.dois.split(",")
        if self.pmids:
            value_dict["pmids"] = [int(x) for x in self.pmids.split(",")]
        if self.biogrid_ids:
            value_dict["biogrid_ids"] = [int(x) for x in self.biogrid_ids.split(",")]
        return value_dict

    def get_pmod_as_value_dict(self):
        """Get pmod metadata as dictionary."""
        return {
            "bel": f"pmod({self.modConfig.pmod_bel})",
            "type": normalized_pmod_reverse[self.modConfig.pmod_bel],
        }

    @property
    def edge_name(self):
        """Get edge name."""
        return self.modConfig.effect + "_" + normalized_pmod_reverse[self.modConfig.pmod_bel] + "_bg"


class Effect:
    """Class definition for Effect."""

    INCREASES = "increases"
    DECREASES = "decreases"


class ModConfig:
    """Modification configuration."""

    def __init__(self, bg_mod_name: str, effect: str, pmod_bel: str = None):
        """Init method for ModConfig class.

        Parameters
        ----------
        bg_mod_name: str
            BioGrid modification name
        effect: str
            Effect of BioGrid Edge
        pmod_bel: str
            Posttranslational modification as BEL
        """
        self.bg_mod_name: str = bg_mod_name
        self.effect: str = effect
        self.pmod_bel: str = pmod_bel


class Modification(Enum):
    """BioGrid modification and configuration for BEL converting."""

    PHOSPHORYLATION: ModConfig = ModConfig("Phosphorylation", Effect.INCREASES, BelPmod.PHO)
    UBIQUITINATION: ModConfig = ModConfig("Ubiquitination", Effect.INCREASES, BelPmod.UBI)
    ACETYLATION: ModConfig = ModConfig("Acetylation", Effect.INCREASES, BelPmod.ACE)
    DEUBIQUITINATION: ModConfig = ModConfig("Deubiquitination", Effect.DECREASES, BelPmod.UBI)
    PROTEOLYTIC_PROCESSING: ModConfig = ModConfig("Proteolytic Processing", Effect.DECREASES)
    METHYLATION: ModConfig = ModConfig("Methylation", Effect.INCREASES, BelPmod.ME0)
    SUMOYLATION: ModConfig = ModConfig("Sumoylation", Effect.INCREASES, BelPmod.SUM)
    DEPHOSPHORYLATION: ModConfig = ModConfig("Dephosphorylation", Effect.DECREASES, BelPmod.PHO)
    DEACETYLATION: ModConfig = ModConfig("Deacetylation", Effect.DECREASES, BelPmod.ACE)
    NEDD_RUB1_YLATION: ModConfig = ModConfig("Nedd(Rub1)ylation", Effect.INCREASES, BelPmod.NED)
    RIBOSYLATION: ModConfig = ModConfig("Ribosylation", Effect.INCREASES, BelPmod.ADD)
    DESUMOYLATION: ModConfig = ModConfig("Desumoylation", Effect.DECREASES, BelPmod.SUM)
    DEMETHYLATION: ModConfig = ModConfig("Demethylation", Effect.DECREASES, BelPmod.ME0)
    DENEDDYLATION: ModConfig = ModConfig("Deneddylation", Effect.DECREASES, BelPmod.NED)
    PRENYLATION: ModConfig = ModConfig("Prenylation", Effect.INCREASES, BelPmod.PRE)
    GLYCOSYLATION: ModConfig = ModConfig("Glycosylation", Effect.INCREASES, BelPmod.GLY)
    NEDDYLATION: ModConfig = ModConfig("Neddylation", Effect.INCREASES, BelPmod.NED)
    DE_ISGYLATION: ModConfig = ModConfig("de-ISGylation", Effect.INCREASES, BelPmod.DEI)
    FAT10YLATION: ModConfig = ModConfig("FAT10ylation", Effect.INCREASES, BelPmod.FAT)
    UFMYLATION: ModConfig = ModConfig("Ufmylation", Effect.INCREASES, BelPmod.UFM)
    ISGYLATION: ModConfig = ModConfig("ISGylation", Effect.INCREASES, BelPmod.ISG)

    @classmethod
    def get_reverse_dict(cls) -> typing.Dict[str, ModConfig]:
        """Reverse the key and values for the dictionary."""
        return {cls.__members__[x].value.bg_mod_name: cls.__members__[x].value for x in cls.__members__}


MODIFICATIONS = tuple(Modification.get_reverse_dict().keys())


class BioGrid(odb_meta.Graph):
    """Class definition for BioGRID."""

    def __init__(self, client: OrientDB = None):
        """Biogrid database."""
        self.client = client
        self.biodb_name = BIOGRID
        self.url = urls.BIOGRID
        self.urls = {self.biodb_name: self.url}
        super().__init__(
            edges=odb_structure.biogrid_edges,
            urls=self.urls,
            tables_base=biogrid.Base,
            biodb_name=self.biodb_name,
        )
        self.file_path = tools.get_file_path(urls.BIOGRID, self.biodb_name)

        self.bel_rid_cache = {}

    def __len__(self) -> int:
        """Get number of 'biogrid_interaction' graph edges."""
        return self.number_of_edges["biogrid_interaction"]

    def __contains__(self, biogrid_id) -> bool:
        """Check if biogrid_interaction edge with biogrid_id exists in graph."""
        # TODO: Check if this is still valid
        sql = "Select 1 from biogrid_interaction where any(biogrid).biogrid_id = {}".format(biogrid_id)
        return bool(len(self.execute(sql)))

    def insert_data(self) -> Dict[str, int]:
        """Insert BioGRID data into database."""
        use_columns = {
            "#BioGRID Interaction ID": "biogrid_id",
            "BioGRID ID Interactor A": "biogrid_a_id",
            "BioGRID ID Interactor B": "biogrid_b_id",
            "Entrez Gene Interactor A": "entrez_a",
            "Entrez Gene Interactor B": "entrez_b",
            "Systematic Name Interactor A": "systematic_name_a",
            "Systematic Name Interactor B": "systematic_name_b",
            "Official Symbol Interactor A": "symbol_a",
            "Official Symbol Interactor B": "symbol_b",
            "Experimental System": "experimental_system",
            "Experimental System Type": "experimental_system_type",
            "Author": "author",
            "Publication Source": "publication_source",
            "Organism ID Interactor A": "taxonomy_a_id",
            "Organism ID Interactor B": "taxonomy_b_id",
            "Throughput": "throughput",
            "Score": "score",
            "Modification": "modification",
            "Qualifications": "qualification",
            "Source Database": "source",
            "SWISS-PROT Accessions Interactor A": "uniprot_a",
            "TREMBL Accessions Interactor A": "trembl_a",
            "SWISS-PROT Accessions Interactor B": "uniprot_b",
            "TREMBL Accessions Interactor B": "trembl_b",
            "Organism Name Interactor A": "org_a",
            "Organism Name Interactor B": "org_b",
        }

        # main table
        df = pd.read_csv(self.file_path, usecols=use_columns.keys(), sep="\t", low_memory=False)
        df.rename(columns=use_columns, inplace=True)
        df.replace("-", np.nan, inplace=True)

        # experimental system
        df = self._create_experimental_system_table(df)

        # taxonomy
        self._create_taxonomy_table(df)

        # source
        df = self._create_source_table(df)

        # interactor
        df = self._create_interactor_table(df)

        # modification
        df = self._create_modification_table(df)

        # throughput
        df = self._create_throughput_table(df)

        # publication
        df = self._create_publication_table(df)

        # save main
        df.index += 1
        df.index.rename("id", inplace=True)

        df.to_sql(biogrid.Biogrid.__tablename__, self.engine, if_exists="append")

        return {self.biodb_name: df.shape[0]}

    def _create_publication_table(self, df: pd.DataFrame) -> pd.DataFrame:
        df_ay = df.author.str.extract(r"^(?P<author_name>[^(]+)\s*\((?P<publication_year>\d+)\)$")
        df_source = df.publication_source.str.extract(r"^(?P<source>[^:]+):(?P<source_identifier>.*)")
        df_pub = pd.concat([df_ay, df_source, df.publication_source], axis=1)
        df_pub.drop_duplicates(inplace=True)
        df_pub.reset_index(inplace=True)
        df_pub.index += 1
        df_pub.index.rename("id", inplace=True)
        df_pub["publication_id"] = df_pub.index
        df_pub_4join = df_pub.set_index("publication_source")[["publication_id"]]
        df = (
            df_pub_4join.join(df.set_index("publication_source"), how="right")
            .reset_index()
            .drop(columns=["publication_source", "author"])
        )
        df_pub.drop(columns=["publication_source", "index", "publication_id"], inplace=True)
        df_pub.to_sql(biogrid.Publication.__tablename__, self.engine, if_exists="append")

        return df

    def _create_modification_table(self, df: pd.DataFrame) -> pd.DataFrame:
        df_mod = df[["modification"]].dropna().value_counts().reset_index().rename(columns={0: "frequency"})
        df_mod.index += 1
        df_mod.index.rename("id", inplace=True)
        df_mod["modification_id"] = df_mod.index
        df = (
            df_mod.set_index("modification")[["modification_id"]]
            .join(df.set_index("modification"), how="right")
            .reset_index()
            .drop(columns=["modification"])
        )
        df_mod.drop(columns=["modification_id"]).to_sql(
            biogrid.Modification.__tablename__, self.engine, if_exists="append"
        )
        return df

    def _create_throughput_table(self, df: pd.DataFrame) -> pd.DataFrame:
        df_tp = df[["throughput"]].value_counts().reset_index().rename(columns={0: "frequency"})
        df_tp.index += 1
        df_tp.index.rename("id", inplace=True)
        df_tp["throughput_id"] = df_tp.index
        df_tp_join = df_tp.set_index("throughput")[["throughput_id"]]
        df = df_tp_join.join(df.set_index("throughput"), how="right").reset_index().drop(columns=["throughput"])
        df_tp.drop(columns=["throughput_id"], inplace=True)
        df_tp.to_sql(biogrid.Throughput.__tablename__, self.engine, if_exists="append")
        return df

    def _create_interactor_table(self, df: pd.DataFrame) -> pd.DataFrame:
        columns = [
            "entrez",
            "biogrid_id",
            "systematic_name",
            "symbol",
            "uniprot",
            "trembl",
            "taxonomy_id",
        ]
        cols_a = [
            "entrez_a",
            "biogrid_a_id",
            "systematic_name_a",
            "symbol_a",
            "uniprot_a",
            "trembl_a",
            "taxonomy_a_id",
        ]
        cols_b = [
            "entrez_b",
            "biogrid_b_id",
            "systematic_name_b",
            "symbol_b",
            "uniprot_b",
            "trembl_b",
            "taxonomy_b_id",
        ]
        df_a = df[cols_a]
        df_a.columns = columns
        df_b = df[cols_b]
        df_b.columns = columns
        df_ia = pd.concat([df_a.set_index("biogrid_id"), df_b.set_index("biogrid_id")]).drop_duplicates()
        # extract the first accession
        df_ia.uniprot = df_ia.uniprot.str.split("|").str[0]
        df_ia.trembl = df_ia.trembl.str.split("|").str[0]
        df_ia.replace("-", None).to_sql(biogrid.Interactor.__tablename__, self.engine, if_exists="append")
        cols4delete = list(set(cols_a + cols_b) - {"biogrid_a_id", "biogrid_b_id"})
        df.drop(columns=cols4delete, inplace=True)
        return df

    def _create_experimental_system_table(self, df: pd.DataFrame) -> pd.DataFrame:
        # extract ExperimentalSystem
        df_exp = (
            df[["experimental_system", "experimental_system_type"]]
            .value_counts()
            .reset_index()
            .rename(columns={0: "frequency"})
        )
        df_exp.index += 1
        df_exp.index.rename("id", inplace=True)
        df_exp["experimental_system_id"] = df_exp.index
        # join ExperimentalSystem ID to main table and drop cols experimental_system, experimental_system_type
        df = (
            df_exp.set_index(["experimental_system", "experimental_system_type"])[["experimental_system_id"]]
            .join(df.set_index(["experimental_system", "experimental_system_type"]))
            .reset_index()
            .drop(columns=["experimental_system", "experimental_system_type"])
        )
        # save ExperimentalSystem
        df_exp.drop(columns=["experimental_system_id"]).to_sql(
            biogrid.ExperimentalSystem.__tablename__, self.engine, if_exists="append"
        )
        return df

    def _create_taxonomy_table(self, df: pd.DataFrame):
        df_org_a = df[["taxonomy_a_id", "org_a"]].rename(
            columns={"taxonomy_a_id": "taxonomy_id", "org_a": "organism_name"}
        )
        df_org_b = df[["taxonomy_b_id", "org_b"]].rename(
            columns={"taxonomy_b_id": "taxonomy_id", "org_b": "organism_name"}
        )
        df_taxonomy = pd.concat([df_org_a, df_org_b])
        df_taxonomy.drop_duplicates(inplace=True)
        df_taxonomy.reset_index(inplace=True)
        df_taxonomy.drop(columns=["index"], inplace=True)
        df_taxonomy.set_index("taxonomy_id", inplace=True)
        df_taxonomy.to_sql(biogrid.Taxonomy.__tablename__, self.engine, if_exists="append")
        df.drop(columns=["org_a", "org_b"], inplace=True)

    def _create_source_table(self, df: pd.DataFrame) -> pd.DataFrame:
        df_source = df[["source"]].drop_duplicates().reset_index().drop(columns=["index"])
        df_source.index += 1
        df_source.index.rename("id", inplace=True)
        df_source["source_id"] = df_source.index
        df = df_source.set_index(["source"]).join(df.set_index(["source"])).reset_index().drop(columns=["source"])
        df_source.drop(columns=["source_id"]).to_sql(biogrid.Source.__tablename__, self.engine, if_exists="append")
        return df

    def get_uniprot_modification_pairs(self):
        """Return all UniProt modification pairs."""
        # TODO: sql as sqlalchemy query
        sql = """Select
            ia.symbol as subject_symbol,
            ia.uniprot as subject_uniprot,
            ia.taxonomy_id as subject_taxonomy_id,
            ib.symbol as object_symbol,
            ib.uniprot as object_uniprot,
            ib.taxonomy_id as object_taxonomy_id
        from
            biogrid b
            inner join biogrid_interactor ia on (b.biogrid_a_id=ia.biogrid_id)
            inner join biogrid_interactor ib on (b.biogrid_b_id=ib.biogrid_id)
            inner join biogrid_modification m on (m.id=b.modification_id)
        where
            m.modification != 'No Modification' and ia.uniprot IS NOT NULL and ib.uniprot IS NOT NULL
        group by
            ia.symbol,
            ia.uniprot,
            ia.taxonomy_id,
            ib.symbol,
            ib.uniprot,
            ib.taxonomy_id"""
        return [dict(x) for x in self.engine.execute(sql).fetchall()]

    def get_create_pure_protein_rid_by_uniprot(self, taxonomy_id, symbol, uniprot):
        """Get pure protein rid by UniProt accession ID if the protein is involved in a BEL statement."""
        namespace = STANDARD_NAMESPACES.get(taxonomy_id, "UNIPROT")
        name = uniprot if namespace == "UNIPROT" else symbol
        bel = f'p({namespace}:"{name}")'

        if bel in self.bel_rid_cache:
            rid = self.bel_rid_cache[bel]
        else:
            value_dict = {
                "bel": bel,
                "pure": True,
                "namespace": namespace,
                "name": name,
            }
            rid = self.get_create_rid("protein", value_dict=value_dict, check_for="bel")
            self.bel_rid_cache[bel] = rid
        return rid

    def update_interactions(self) -> int:
        """Updates all BioGrid interactions."""
        # TODO: sql_temp as sqlalchemy query
        sql_temp = """
        Select
            ia.symbol as subject_symbol,
            ia.uniprot as subject_uniprot,
            ia.taxonomy_id as subject_taxonomy_id,
            m.modification,
            ib.symbol as object_symbol,
            ib.uniprot as object_uniprot,
            ib.taxonomy_id as object_taxonomy_id,
            es.experimental_system,
            group_concat( distinct b.biogrid_id) as biogrid_ids,
            group_concat( distinct if(p.source='PUBMED',CAST(p.source_identifier AS UNSIGNED),NULL)) as pmids,
            count(distinct p.source_identifier) as num_pubs,
            group_concat( distinct if(p.source='DOI',CAST(p.source_identifier AS UNSIGNED),NULL)) as dois
        from
            biogrid b
            inner join biogrid_interactor ia on (b.biogrid_a_id=ia.biogrid_id)
            inner join biogrid_interactor ib on (b.biogrid_b_id=ib.biogrid_id)
            inner join biogrid_modification m on (m.id=b.modification_id)
            inner join biogrid_publication p on (b.publication_id=p.id)
            inner join biogrid_experimental_system es on (b.experimental_system_id=es.id)
        where
            (ia.uniprot = '{subject_uniprot}' and ib.uniprot = '{object_uniprot}') and
            m.modification != 'No Modification'
        group by
            ia.symbol,
            ia.uniprot,
            ia.taxonomy_id,
            m.modification,
            ib.symbol,
            ib.uniprot,
            ib.taxonomy_id,
            es.experimental_system"""

        uniprots_in_bel_set = self.get_pure_uniprots_in_bel_context()
        uniprot_modification_pairs = self.get_uniprot_modification_pairs()

        counter = 0
        self.clear_edges()

        for e in tqdm(
            uniprot_modification_pairs,
            desc=f"Update {self.biodb_name.upper()} interactions",
        ):
            if e["subject_uniprot"] in uniprots_in_bel_set or e["object_uniprot"] in uniprots_in_bel_set:
                subj_pure_rid = self.get_create_pure_protein_rid_by_uniprot(
                    taxonomy_id=e["subject_taxonomy_id"],
                    symbol=e["subject_symbol"],
                    uniprot=e["subject_uniprot"],
                )

                obj_pure_rid = self.get_create_pure_protein_rid_by_uniprot(
                    taxonomy_id=e["object_taxonomy_id"],
                    symbol=e["object_symbol"],
                    uniprot=e["object_uniprot"],
                )

                sql = sql_temp.format(
                    subject_uniprot=e["subject_uniprot"],
                    object_uniprot=e["object_uniprot"],
                )

                for row in self.engine.execute(sql).fetchall():
                    row_dict = dict(row)
                    be = BioGridEdge(subject_rid=subj_pure_rid, object_rid=obj_pure_rid, **row_dict)
                    edge_value_dict = be.get_edge_value_dict()

                    if be.modConfig.bg_mod_name == "Proteolytic Processing":
                        self.create_edge(
                            "decreases_bg",
                            from_rid=subj_pure_rid,
                            to_rid=obj_pure_rid,
                            value_dict=edge_value_dict,
                        )
                        counter += 1
                    else:
                        obj_pmod_value_dict = be.obj.get_pmod_protein_as_value_dict()
                        pmod_protein_rid = self.node_exists("protein", obj_pmod_value_dict, check_for="bel")
                        if not pmod_protein_rid:
                            pmod_protein_rid = self.get_create_rid("protein", obj_pmod_value_dict, check_for="bel")
                            self.create_edge("has_modified_protein", obj_pure_rid, pmod_protein_rid)
                            pmod_rid = self.insert_record("pmod", be.get_pmod_as_value_dict())
                            self.create_edge("has__pmod", pmod_protein_rid, pmod_rid)
                        self.create_edge(
                            be.edge_name,
                            subj_pure_rid,
                            pmod_protein_rid,
                            edge_value_dict,
                        )
                        counter += 1
        return counter

    def create_view(self):
        """Create SQL view of BioGRID data."""
        sql = """create view if not exists biogrid_view as
            select
                b.biogrid_id,
                ia.symbol as symbol_a,
                ia.uniprot as uniprot_a,
                ta.taxonomy_id as tax_id_a,
                ta.organism_name as organism_a,
                ib.symbol as symbol_b,
                ib.uniprot as uniprot_b,
                tb.taxonomy_id as tax_id_b,
                tb.organism_name as organism_b,
                es.experimental_system,
                m.modification,
                s.source,
                b.qualification,
                p.source as publication_source,
                p.source_identifier as publication_identifier
            from
                biogrid b inner join
                biogrid_interactor ia on (ia.biogrid_id=b.biogrid_a_id) inner join
                biogrid_interactor ib on (ib.biogrid_id=b.biogrid_b_id) inner join
                biogrid_taxonomy ta on (ia.taxonomy_id=ta.taxonomy_id) inner join
                biogrid_taxonomy tb on (ib.taxonomy_id=tb.taxonomy_id) left join
                biogrid_experimental_system es on (b.experimental_system_id=es.id) left join
                biogrid_modification m on (m.id=b.modification_id) left join
                biogrid_source s on (s.id=b.source_id) left join
                biogrid_publication p on (p.id=b.publication_id)"""
        self.engine.execute(sql)



"""UniProt module."""
import ftplib
import logging
import os
import re
from collections import namedtuple
from typing import Dict, List, Tuple, Union

import pandas as pd
from lxml.etree import iterparse
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.defaults import default_tax_ids
from ebel.manager.orientdb import odb_meta, urls
from ebel.manager.orientdb.constants import UNIPROT
from ebel.manager.rdbms.models import uniprot as up
from ebel.tools import get_file_path, gunzip

logger = logging.getLogger(__name__)

Xrefs = Dict[Tuple[str, str], int]
Functions = Dict[str, int]
ScLocations = Dict[str, int]
Organisms = Dict[int, str]
Keywords = Dict[int, str]

Xpath = namedtuple(
    "Xpath",
    (
        "recommended_name",
        "gene_name",
        "organism_scientific",
        "taxid",
        "function",
        "subcellular_locations",
        "location",
    ),
)

XML_URL = "http://uniprot.org/uniprot"
XML_NAMESPACE = f"{{{XML_URL}}}"
XN = {"n": XML_URL}


class UniProtEntry:
    """Helper class for the import of data into RDMS."""

    column_names = [
        "name",
        "accession",
        "recommended_name",
        "gene_names",
        "taxid",
        "host_taxids",
        "function_id",
        "xref_ids",
        "keyword_ids",
        "subcellular_location_ids",
    ]

    def __init__(self):
        """Init UniProt."""
        self.name: Union[str, None] = None
        self.accession: Union[str, None] = None
        self.recommended_name: Union[str, None] = None
        self.gene_names: List[str] = []
        self.taxid: Union[int, None] = None
        self.host_taxids: List[int] = []
        self.function_id: Union[int, None] = None
        self.xref_ids: List[int] = []
        self.keyword_ids: List[int] = []
        self.subcellular_location_ids: List[int] = []

    def get_data_dict(self):
        """Return dict with column names as keys and their associated value tuples as values."""
        return dict(zip(self.column_names, self.get_data_tuple()))

    def get_data_tuple(self):
        """Return a tuple containing metadata for an entry."""
        return (
            self.name,
            self.accession,
            self.recommended_name,
            self.gene_names,
            self.taxid,
            self.host_taxids,
            self.function_id,
            self.xref_ids,
            self.keyword_ids,
            self.subcellular_location_ids,
        )


class UniProt(odb_meta.Graph):
    """UniProt Importer and interface."""

    columns2import = [
        "dataset",
        "accession",
        "name",
        "protein",
        "gene",
        "comment",
        "proteinExistence",
        "feature",
        "evidence",
        "dbReference",
    ]

    def __init__(self, client: OrientDB = None, tax_ids: list = default_tax_ids):
        """Init UniProt."""
        self.client = client
        self.biodb_name = UNIPROT
        self.file_path = get_file_path(urls.UNIPROT_SPROT, self.biodb_name)
        self.file_path_gunzipped = self.file_path.split(".gz")[0]
        super().__init__(
            urls={UNIPROT: urls.UNIPROT_SPROT},
            tables_base=up.Base,
            biodb_name=self.biodb_name,
        )
        self.tax_ids = tax_ids

        self.xpath_pattern = Xpath(
            recommended_name="./n:recommendedName/n:fullName[1]/text()",
            gene_name="./n:name/text()",
            organism_scientific="./n:name[@type='scientific'][1]/text()",
            taxid="./n:dbReference[@type='NCBI Taxonomy'][1]/@id",
            function="./n:text/text()",
            subcellular_locations="./n:subcellular_location/n:location/text()",
            location="./n:subcellularLocation/n:location",
        )

    def __contains__(self, accession) -> bool:
        """Check if protein node with accession exists in graph."""
        results = self.execute(f"Select 1 from protein where uniprot = '{accession}'")
        return bool(len(results))

    def __len__(self) -> int:
        """Return number of proteins (nodes) with uniprot accession number."""
        sql = "Select count(*) from protein where uniprot IS NOT NULL"
        return self.execute(sql)[0].oRecordData["count"]

    def insert_data(self) -> Dict[str, int]:
        """Insert UniProt data depending on NCBI taxonomy identifier."""
        dialect = self.session.bind.dialect.name
        if dialect == "mysql":
            self.engine.execute("SET FOREIGN_KEY_CHECKS=0")

        inserted = self.insert_uniprot()
        self.add_gene_symbols()
        self.session.commit()

        if dialect == "mysql":
            self.engine.execute("SET FOREIGN_KEY_CHECKS=1")

        return {self.biodb_name: inserted}

    def add_gene_symbols(self):
        """Add gene symbols to all tables."""
        self._add_hgnc_symbols()
        self._add_mgi_symbols()
        self._add_rgd_symbols()
        self._add_fly_symbols()

    def _delete_gene_symbols(self, taxid) -> None:
        self.session.query(up.GeneSymbol).filter(up.GeneSymbol.uniprot.has(up.Uniprot.taxid == taxid)).delete(
            synchronize_session=False
        )
        self.session.commit()

    def _add_hgnc_symbols(self) -> int:
        """Add HGNC symbols from genenames.org."""
        inserted = 0
        self._delete_gene_symbols(9606)

        df = pd.read_csv(urls.HGNC_TSV, sep="\t", low_memory=False, usecols=["symbol", "uniprot_ids"]).dropna()

        for row in tqdm(
            df.itertuples(index=False),
            desc="Add HGNC symbols to UniProt",
            total=df.shape[0],
        ):
            uniprot = self.session.query(up.Uniprot).filter(up.Uniprot.accession == row.uniprot_ids).scalar()
            if uniprot:
                self.session.add(up.GeneSymbol(symbol=row.symbol, uniprot_id=uniprot.id))
                inserted += 1
        self.session.commit()

        return inserted

    def _add_mgi_symbols(self) -> int:
        """Add MGI symbols from genenames.org."""
        inserted = 0

        self._delete_gene_symbols(10090)

        df = pd.read_csv(
            urls.UNIPROT_MGI,
            sep="\t",
            usecols=[1, 6],
            names=["symbol", "uniprot_accession"],
        ).dropna()

        for row in tqdm(
            df.itertuples(index=False),
            desc="Add MGI symbols to UniProt",
            total=df.shape[0],
        ):
            uniprot = self.session.query(up.Uniprot).filter(up.Uniprot.accession == row.uniprot_accession).scalar()

            if uniprot:
                self.session.add(up.GeneSymbol(symbol=row.symbol, uniprot_id=uniprot.id))
                inserted += 1
        self.session.commit()

        return inserted

    def _add_rgd_symbols(self) -> int:
        """Add MGI symbols from genenames.org."""
        inserted = 0

        self._delete_gene_symbols(10116)

        df = pd.read_csv(
            urls.UNIPROT_RGD,
            sep="\t",
            comment="#",
            low_memory=False,
            usecols=["SYMBOL", "UNIPROT_ID"],
        ).dropna()
        df.columns = ["symbol", "uniprot_accessions"]
        df.uniprot_accessions = df.uniprot_accessions.str.split(";")

        for row in tqdm(
            df.itertuples(index=False),
            desc="Add RGD symbols to UniProt",
            total=df.shape[0],
        ):
            for accession in row.uniprot_accessions:
                uniprot = self.session.query(up.Uniprot).filter(up.Uniprot.accession == accession).scalar()
                if uniprot:
                    self.session.add(up.GeneSymbol(symbol=row.symbol, uniprot_id=uniprot.id))
                    inserted += 1
        self.session.commit()

        return inserted

    def _add_fly_symbols(self) -> int:
        """Add MGI symbols from genenames.org."""
        inserted = 0
        ftp_base = "ftp.flybase.org"
        ftp_folder = "/releases/current/precomputed_files/genes/"
        ftp = ftplib.FTP(ftp_base)
        ftp.login("anonymous", "anonymous")
        files = ftp.nlst(ftp_folder)
        file_found = [x for x in files if re.search(r"fbgn_NAseq_Uniprot_fb_\d{4}_\d{2}.tsv.gz", x)]
        if file_found:
            url = f"ftp://{ftp_base}{file_found[0]}"
            inserted = 0

            self._delete_gene_symbols(7227)

            df = pd.read_csv(
                url,
                comment="#",
                sep="\t",
                header=None,
                usecols=[0, 5],
                names=["symbol", "uniprot_accession"],
            ).dropna()

            for row in tqdm(
                df.itertuples(index=False),
                desc="Add FlyBase symbols to UniProt",
                total=df.shape[0],
            ):
                uniprot = self.session.query(up.Uniprot).filter(up.Uniprot.accession == row.uniprot_accession).scalar()
                if uniprot:
                    self.session.add(up.GeneSymbol(symbol=row.symbol, uniprot_id=uniprot.id))
                    inserted += 1
            self.session.commit()
        ftp.close()
        return inserted

    def update_bel(self) -> Dict[str, int]:
        """Update uniprot links in protein nodes to uniprot class."""
        updated = dict()
        updated["HGNC"] = self._update_proteins("HGNC", 9606)
        updated["MGI"] = self._update_proteins("MGI", 10090)
        updated["RGD"] = self._update_proteins("RGD", 10116)
        updated["UNIPROT"] = self._update_uniprot_proteins()
        return updated

    def _update_protein_node(self, uniprot_accession, recommended_name, name, namespace, taxid) -> int:
        """Update all proteins using UNIPROT as namespace. Returns numbers of updated."""
        sql = (
            f'Update protein set uniprot = "{uniprot_accession}", label = "{recommended_name}", '
            f'species = {taxid} where namespace = "{namespace}" and name = "{name}"'
        )
        return self.execute(sql)[0]

    def _get_accesssion_recname(self, taxid, gene_symbol) -> Union[Tuple[str, str], None]:
        """Get uniprot accession and recommended name from HGNC.

        If this has no result it tries uniprot by gene symbol and NCBI taxonomy ID.
        """
        # TODO: This is in general a dangerous method because it selects the first accession number, but there could
        # be more than one
        sql = (
            f"Select accession, recommended_name from uniprot as u inner join uniprot_gene_symbol as gs "
            f'on (u.id=gs.uniprot_id) where u.taxid={taxid} and gs.symbol="{gene_symbol}" limit 1'
        )
        results = self.engine.execute(sql)
        return results.fetchone() if results else None

    def _update_proteins(self, namespace, taxid) -> int:
        """Update human proteins (namespace HGNC) with uniprot accession and recommended name.

        Returns number of updated proteins.
        """
        sql = f'SELECT distinct(name) as name from protein WHERE namespace="{namespace}"'
        updated = 0
        for protein in self.query(sql).itertuples(index=False):
            acc_rec = self._get_accesssion_recname(taxid, protein.name)
            if acc_rec:
                accession, recommended_name = acc_rec
                num_updated = self._update_protein_node(accession, recommended_name, protein.name, namespace, taxid)
                updated += num_updated
            else:
                err_txt = (
                    f"NO MAPPING TO UNIPROT: {namespace}:{protein.name} can not be mapped to uniprot. "
                    f"Possible reasons: Withdrawn or only predicted."
                )
                logging.warning(err_txt)
        return updated

    def _get_recname_taxid_by_accession_from_uniprot_api(self, accession) -> Tuple[str, int]:
        """Fetch uniprot entry by accession and adds to the database. Returns recommended name."""
        sql = f"Select recommended_name,taxid from uniprot where accession='{accession}' limit 1"
        result = self.engine.execute(sql).fetchone()
        if result:
            return result

    def _update_uniprot_proteins(self) -> int:
        """Update all proteins using UNIPROT as namespace. Returns number of updated proteins."""
        updated = 0
        sql_temp = "Select recommended_name, taxid from uniprot where accession='{}' limit 1"
        sql_uniprot = 'SELECT distinct(name) as accession from protein WHERE namespace="UNIPROT"'
        sql_update = (
            'Update protein set uniprot = name, label = "{}", species = {} '
            'where namespace = "UNIPROT" and name = "{}"'
        )
        for protein in self.query(sql_uniprot).itertuples(index=False):
            sql = sql_temp.format(protein.accession)
            found = self.engine.execute(sql).fetchone()
            if found:
                recommended_name, taxid = found
                num_updated = self.execute(sql_update.format(recommended_name, taxid, protein.accession))[0]
                updated += num_updated
            else:
                recname_taxid = self._get_recname_taxid_by_accession_from_uniprot_api(protein.accession)
                if recname_taxid:
                    recommended_name, taxid = recname_taxid
                    num_updated = self.execute(sql_update.format(recommended_name, taxid, protein.accession))[0]
                    updated += num_updated
        return updated

    def __read_linked_tables(
        self,
    ) -> Tuple[Keywords, Organisms, Xrefs, Functions, ScLocations, int]:
        """Return xref, functions and subcellular locations."""
        xrefs: Xrefs = {}
        functions: Functions = {}
        sclocations: ScLocations = {}
        organisms: Organisms = {}
        xref_index = 0
        function_index = 0
        sclocation_index = 0
        keywords: Keywords = {}

        doc = iterparse(self.file_path_gunzipped, events=("end",), tag=f"{XML_NAMESPACE}entry")

        counter = 0

        logger.info("Read linked UniProt tables")

        for action, elem in tqdm(doc, desc="Get linked UniProt data"):
            counter += 1

            for child in elem.iterchildren():
                ctag = child.tag[len(XML_NAMESPACE) :]

                if ctag == "dbReference":
                    x_type_id = (child.attrib.get("type"), child.attrib.get("id"))
                    if x_type_id not in xrefs:
                        xref_index += 1
                        xrefs[x_type_id] = xref_index

                elif ctag in ["organismHost", "organism"]:
                    taxid = int(child.xpath(self.xpath_pattern.taxid, namespaces=XN)[0])
                    if taxid not in organisms:
                        organism_scientific = child.xpath(self.xpath_pattern.organism_scientific, namespaces=XN)[0]
                        organisms[taxid] = organism_scientific

                elif ctag == "keyword":
                    keyword_id = int(child.attrib.get("id").split("KW-")[1])
                    if keyword_id not in keywords:
                        keywords[keyword_id] = child.text

                elif ctag == "comment":
                    ctype = child.attrib.get("type")

                    if ctype == "function":
                        func = child.xpath(self.xpath_pattern.function, namespaces=XN)[0]

                        if func not in functions:
                            function_index += 1
                            functions[func] = function_index

                    elif ctype == "subcellular location":
                        scls = child.findall(self.xpath_pattern.location, namespaces=XN)

                        for scl in [x.text for x in scls]:
                            if scl not in sclocations:
                                sclocation_index += 1
                                sclocations[scl] = sclocation_index
            elem.clear()
        return keywords, organisms, xrefs, functions, sclocations, counter

    def __insert_linked_data(
        self,
        keywords: Keywords,
        organisms: Organisms,
        xrefs: Xrefs,
        functions: Functions,
        sclocations: ScLocations,
    ):
        """Insert data from file."""
        if keywords:
            df_kw = pd.DataFrame(keywords.items(), columns=["keywordid", "keyword_name"])
            self._insert_into_database(df_kw, up.Keyword)

        if organisms:
            df_org = pd.DataFrame(organisms.items(), columns=["taxid", "scientific_name"])
            self._insert_into_database(df_org, up.Organism)

        if sclocations:
            df_sl = pd.DataFrame([x[::-1] for x in sclocations.items()], columns=["id", "name"])
            self._insert_into_database(df_sl, up.SubcellularLocation)

        if xrefs:
            df_xr = pd.DataFrame(
                [(v, k[0], k[1]) for k, v in xrefs.items()],
                columns=["id", "db", "identifier"],
            )
            self._insert_into_database(df_xr, up.Xref)

        if functions:
            df_fc = pd.DataFrame([x[::-1] for x in functions.items()], columns=["id", "description"])
            self._insert_into_database(df_fc, up.Function)

    def _insert_into_database(self, dataframe, model):
        logger.info(f"insert into {model.__tablename__}")
        dataframe.to_sql(
            model.__tablename__,
            self.engine,
            chunksize=100000,
            if_exists="append",
            index=False,
        )
        inserted = dataframe.shape[0]
        del dataframe
        return inserted

    def insert_uniprot(self) -> int:
        """Insert UniProt data into database."""
        logger.info("Drop and create Uniprot table in RDBMS")

        logger.info("Insert data linked to Uniprot entry into RDBMS")
        # avoid to use old gunzipped file
        if os.path.exists(self.file_path_gunzipped):
            os.remove(self.file_path_gunzipped)
        if not os.path.exists(self.file_path_gunzipped):
            gunzip(self.file_path, self.file_path_gunzipped)

        (
            keywords,
            hosts,
            xrefs,
            functions,
            sclocations,
            number_of_entries,
        ) = self.__read_linked_tables()
        self.__insert_linked_data(keywords, hosts, xrefs, functions, sclocations)
        inserted = self.__insert_uniprot_data(xrefs, functions, sclocations, number_of_entries)

        # save storage space
        if os.path.exists(self.file_path_gunzipped):
            os.remove(self.file_path_gunzipped)

        # return number_of_entries
        return inserted

    @staticmethod
    def shorten_tag(elem):
        """Shorten the given tag."""
        return elem.tag[len(XML_NAMESPACE) :]

    @staticmethod
    def get_tag(name):
        """Get tag based on given name."""
        return f"{XML_NAMESPACE}{name}"

    def __insert_uniprot_data(self, xrefs, functions, sclocations, number_of_entries) -> int:
        doc = iterparse(self.file_path_gunzipped, events=("end",), tag=self.get_tag("entry"))

        counter = 0

        uniprots = []

        for action, elem in tqdm(doc, total=number_of_entries, desc=f"Import {self.biodb_name.upper()}"):
            counter += 1

            uniprot = UniProtEntry()
            for child in elem.iterchildren():
                ctag = self.shorten_tag(child)

                if ctag == "accession" and not uniprot.accession:
                    uniprot.accession = child.text

                elif ctag == "name":
                    uniprot.name = child.text

                elif ctag == "gene":
                    for cchild in child.iterchildren(tag=self.get_tag("name")):
                        uniprot.gene_names.append(cchild.text)

                elif ctag == "organism":
                    for cchild in child.iterchildren(tag=self.get_tag("dbReference")):
                        if cchild.attrib.get("type") == "NCBI Taxonomy":
                            uniprot.taxid = int(cchild.attrib.get("id"))
                            break

                elif ctag == "protein":
                    for cchild in child.iterchildren(tag=self.get_tag("recommendedName")):
                        for ccchild in cchild.iterchildren(tag=self.get_tag("fullName")):
                            uniprot.recommended_name = ccchild.text
                            break

                elif ctag == "organismHost":
                    for cchild in child.iterchildren(tag=self.get_tag("dbReference")):
                        if cchild.attrib.get("type") == "NCBI Taxonomy":
                            uniprot.host_taxids.append(int(cchild.attrib.get("id")))

                elif ctag == "comment":
                    ctype = child.attrib.get("type")

                    if ctype == "function":
                        for cchild in child.iterchildren(tag=self.get_tag("text")):
                            uniprot.function_id = functions[cchild.text]
                            break

                    elif ctype == "subcellular location":
                        for cchild in child.iterchildren(tag=self.get_tag("subcellularLocation")):
                            for ccchild in cchild.iterchildren(tag=self.get_tag("location")):
                                uniprot.subcellular_location_ids.append(sclocations[ccchild.text])

                elif ctag == "dbReference":
                    xref = child.attrib.get("type"), child.attrib.get("id")
                    uniprot.xref_ids.append(xrefs[xref])

                elif ctag == "keyword":
                    keyword_id = int(child.attrib.get("id").split("KW-")[1])
                    uniprot.keyword_ids.append(keyword_id)
            uniprots.append(uniprot)
            elem.clear()

        df = pd.DataFrame([u.get_data_tuple() for u in uniprots], columns=UniProtEntry.column_names)
        df.index += 1
        df.index.rename("id", inplace=True)
        cols_uniprot = ["name", "accession", "recommended_name", "taxid", "function_id"]
        logger.info(f"start insert {up.Uniprot.__tablename__}")
        df[cols_uniprot].to_sql(up.Uniprot.__tablename__, self.engine, if_exists="append", chunksize=100000)

        df["uniprot_id"] = df.index

        logger.info(f"start insert {up.Gene.__tablename__}")
        df[["gene_names", "uniprot_id"]].explode("gene_names").dropna().rename(columns={"gene_names": "name"}).to_sql(
            up.Gene.__tablename__,
            self.engine,
            if_exists="append",
            index=False,
            chunksize=100000,
        )

        logger.info(f"start insert {up.uniprot__uniprot_keyword.name}")
        df[["keyword_ids", "uniprot_id"]].explode("keyword_ids").dropna().rename(
            columns={"keyword_ids": "uniprot_keyword_id"}
        ).to_sql(
            up.uniprot__uniprot_keyword.name,
            self.engine,
            if_exists="append",
            index=False,
            chunksize=100000,
        )

        logger.info(f"start insert {up.uniprot__uniprot_host.name}")
        df[["host_taxids", "uniprot_id"]].explode("host_taxids").dropna().rename(
            columns={"host_taxids": "uniprot_organism_id"}
        ).to_sql(
            up.uniprot__uniprot_host.name,
            self.engine,
            if_exists="append",
            index=False,
            chunksize=100000,
        )

        logger.info(f"start insert {up.uniprot__uniprot_xref.name}")
        df[["xref_ids", "uniprot_id"]].explode("xref_ids").dropna().rename(
            columns={"xref_ids": "uniprot_xref_id"}
        ).to_sql(
            up.uniprot__uniprot_xref.name,
            self.engine,
            if_exists="append",
            index=False,
            chunksize=100000,
        )

        logger.info(f"start insert {up.uniprot__uniprot_subcellular_location.name}")
        df[["subcellular_location_ids", "uniprot_id"]].explode("subcellular_location_ids").dropna().rename(
            columns={"subcellular_location_ids": "uniprot_subcellular_location_id"}
        ).to_sql(
            up.uniprot__uniprot_subcellular_location.name,
            self.engine,
            if_exists="append",
            index=False,
            chunksize=100000,
        )

        return df.shape[0]

    def update_interactions(self) -> int:
        """Abstract method."""
        pass



"""Custom drug2pmod API methods."""

import re
from collections import defaultdict, namedtuple
from math import ceil
from typing import List, Union

from flask import request

from ebel import Bel
from ebel.web.api.ebel.v1 import DataType, OrientDbSqlOperator

SQL_MATCH_TEMPLATE = """
    match {{class:drug, as:drug{drug}}}
    .outE('has_drug_target'){{as: has_drug_target{has_drug_target}}}
    .inV(){{class:protein, as:drug_target{drug_target}}}
    .outE('bel_relation', 'has_ppi_bg', 'has_ppi_kg'){{as:drug_target_to_target{drug_target_to_target}}}
    .inV(){{class:protein, as: target{target}}}-has__pmod->
    {{class:pmod, as:pmod{pmod}}}"""


class Column:
    """Column class."""

    def __init__(
        self,
        odb_class: str,
        form_name: str,
        column: str,
        sql_operator: OrientDbSqlOperator = OrientDbSqlOperator.EQUALS,
        data_type: DataType = DataType.STRING,
    ):
        """Init method."""
        self.odb_class = odb_class
        self.column = column
        self.display_column = column + ".asString()" if "@" in column else column
        self.sql_operator = sql_operator
        self.form_name = form_name
        self.data_type = data_type
        self.value = None

    def set_search_term(self, value: Union[str, None]):
        """Get the value of a given search term."""
        if value is not None and value.strip():
            self.value = value.strip()
            if "%" not in self.value and self.sql_operator == OrientDbSqlOperator.LIKE:
                self.sql_operator = OrientDbSqlOperator.EQUALS


Pagination = namedtuple("Pagination", ["page", "page_size", "skip"])


class Query:
    """Query class for drug2pmod algorithm."""

    def __init__(self, columns: List[Column]):
        """Init method for algorithm."""
        self.columns: List[Column] = columns
        self.ebel = Bel()
        self.wheres_dict: dict = self.get_wheres_dict()

    def get_pagination(self) -> Pagination:
        """Return the results as a collection of paginations."""
        page_size = request.args.get("page_size", "10")
        page_size = int(page_size) if re.search(r"^\d+$", page_size) else 10
        page_size = 10 if page_size >= 100 else page_size
        page = request.args.get("page", "1")
        page = int(page) if re.search(r"^\d+$", page) else 1
        skip = (page - 1) * page_size
        return Pagination(page=page, page_size=page_size, skip=skip)

    def get_wheres_dict(self) -> dict:
        """Generate a generic filter SQL statement."""
        wheres = defaultdict(list)

        for col in self.columns:
            if col.value:
                if col.column.endswith("@rid") and "," in col.value:
                    rids = [x.strip() for x in col.value.split(",") if re.search(r"^#\d+:\d+$", x.strip())]
                    rids_str = "[" + ",".join(rids) + "]"
                    wheres[col.odb_class].append(f"{col.column} in {rids_str}")
                elif col.column != "@class":
                    value = col.value.replace('"', '\\"')

                    if col.data_type in [DataType.STRING, DataType.LIST_STRING]:
                        value = f'"{value}"'
                    else:
                        value = value

                    if col.data_type in [
                        DataType.LIST_STRING,
                        DataType.LIST_NUMBER,
                        DataType.LIST_INTEGER,
                    ]:
                        wheres[col.odb_class].append(f"{value} {col.sql_operator.value} {col.column}")
                    else:
                        wheres[col.odb_class].append(f"{col.column} {col.sql_operator.value} {value}")

        wheres_dict = {odb_class: "" for odb_class in {c.odb_class for c in self.columns}}
        for odb_class, col_queries in wheres.items():
            if col_queries:
                wheres_dict[odb_class] = ", where:( " + " AND ".join(col_queries) + " )"

        return wheres_dict

    @property
    def sql(self) -> str:
        """Create the SQL statement."""
        sql = SQL_MATCH_TEMPLATE.format(**self.wheres_dict)
        return sql

    def get_number_of_results(self) -> int:
        """Count the number of results."""
        nodes_edges = ", ".join({c.odb_class for c in self.columns})
        sql = f"SELECT count(*) FROM ({self.sql} return {nodes_edges})"
        print(sql)
        return self.ebel.query_get_dict(sql)[0]["count"]

    @property
    def __sql4results(self):
        cols = ", ".join([f"{c.odb_class}.{c.display_column} as {c.form_name}" for c in self.columns])
        return self.sql + " return " + cols

    def get_result(self) -> dict:
        """Return a dictionary of result metadata."""
        p = self.get_pagination()

        number_of_results = self.get_number_of_results()
        pages = ceil(number_of_results / p.page_size)

        results = []
        if number_of_results > 0:
            sql_paginated = self.__sql4results + f" skip {p.skip} limit {p.page_size}"
            results = [x for x in self.ebel.query_get_dict(sql_paginated)]

        return {
            "page": p.page,
            "page_size": p.page_size,
            "number_of_results": number_of_results,
            "pages": pages,
            "results": results,
        }


def get_drug2pmod() -> dict:
    """Create a table of relations and information fulfilling the algorithm."""
    columns: List[Column] = [
        Column("drug", "drug__cas_number", "cas_number", OrientDbSqlOperator.LIKE),
        Column("drug", "drug__description", "description", OrientDbSqlOperator.LIKE),
        Column("drug", "drug__drugbank_id", "drugbank_id", OrientDbSqlOperator.LIKE),
        Column("drug", "drug__indication", "indication", OrientDbSqlOperator.LIKE),
        Column("drug", "drug__label", "label", OrientDbSqlOperator.LIKE),
        Column(
            "drug",
            "drug__mechanism_of_action",
            "mechanism_of_action",
            OrientDbSqlOperator.LIKE,
        ),
        Column("drug", "drug__metabolism", "metabolism", OrientDbSqlOperator.LIKE),
        Column(
            "drug",
            "drug__pharmacodynamics",
            "pharmacodynamics",
            OrientDbSqlOperator.LIKE,
        ),
        Column("drug", "drug__toxicity", "toxicity", OrientDbSqlOperator.LIKE),
        Column(
            "has_drug_target",
            "has_drug_target__action",
            "action",
            OrientDbSqlOperator.LIKE,
        ),
        Column(
            "has_drug_target",
            "has_drug_target__known_action",
            "known_action",
            OrientDbSqlOperator.LIKE,
        ),
        Column("drug_target", "drug_target__name", "name", OrientDbSqlOperator.LIKE),
        Column("drug_target", "drug_target__bel", "bel", OrientDbSqlOperator.LIKE),
        Column("drug_target", "drug_target__label", "label", OrientDbSqlOperator.LIKE),
        Column("drug_target", "drug_target__uniprot", "uniprot", OrientDbSqlOperator.LIKE),
        Column(
            "drug_target",
            "drug_target__reactome_pathways",
            "reactome_pathways",
            OrientDbSqlOperator.IN,
            DataType.LIST_STRING,
        ),
        Column(
            "drug_target_to_target",
            "drug_target_to_target__relation",
            "@class",
            OrientDbSqlOperator.EQUALS,
        ),
        Column(
            "drug_target_to_target",
            "drug_target_to_target__evidence",
            "evidence",
            OrientDbSqlOperator.LIKE,
        ),
        Column(
            "drug_target_to_target",
            "drug_target_to_target__pmid",
            "pmid",
            OrientDbSqlOperator.LIKE,
        ),
        Column(
            "drug_target_to_target",
            "drug_target_to_target__citation",
            "citation",
            OrientDbSqlOperator.LIKE,
        ),
        Column("target", "target__name", "name", OrientDbSqlOperator.LIKE),
        Column("target", "target__bel", "bel", OrientDbSqlOperator.LIKE),
        Column("target", "target__label", "label", OrientDbSqlOperator.LIKE),
        Column("target", "target__uniprot", "uniprot", OrientDbSqlOperator.LIKE),
        Column(
            "target",
            "target__reactome_pathways",
            "reactome_pathways",
            OrientDbSqlOperator.IN,
            DataType.LIST_STRING,
        ),
        Column("pmod", "pmod__amino_acid", "amino_acid", OrientDbSqlOperator.LIKE),
        Column("pmod", "pmod__name", "name", OrientDbSqlOperator.LIKE),
        Column("pmod", "pmod__namespace", "namespace", OrientDbSqlOperator.LIKE),
        Column("pmod", "pmod__position", "position", OrientDbSqlOperator.LIKE),
        Column("pmod", "pmod__type", "type", OrientDbSqlOperator.LIKE),
        Column("pmod", "pmod__bel", "bel", OrientDbSqlOperator.LIKE),
    ]

    for column in columns:
        column.set_search_term(request.args.get(column.form_name))

    query = Query(columns)

    return query.get_result()



"""GWAS Catalog API methods."""
from flask import request

from ebel.manager.rdbms.models import gwas_catalog
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_gwas_catalog():
    """Get egenric GWAS catalog entry."""
    return _get_data(gwas_catalog.GwasCatalog)


def get_ebel_relation():
    """Get GWAS Catalog related eBEL relations."""
    allowed_relations = [
        "has_snp_gwascatalog",
        "has_mapped_snp_gc",
        "has_downstream_snp_gc",
        "has_upstream_snp_gc",
    ]
    relation = request.args.get("relation")
    relation = relation if relation in allowed_relations else allowed_relations[0]

    conf = {
        "snp_rid": "in.@rid.asString()",
        "namespace": "out.namespace",
        "name": "out.name",
        "gene_rid": "out.@rid.asString()",
        "pubmed_id": "pubmed_id",
        "disease_trait": "disease_trait",
        "rs_number": "in.rs_number",
        "edge_rid": "@rid.asString()",
    }

    sql = "SELECT "
    sql += ", ".join([f"{v} as {k}" for k, v in conf.items()])
    sql += f" FROM {relation}"

    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}
    if paras:
        sql += " WHERE " + " AND ".join([f'{conf[k]} like "{v}"' for k, v in paras.items()])

    return _get_paginated_ebel_query_result(sql, print_sql=True)



"""IUPHAR API methods."""
from flask.globals import request

from ebel.manager.orientdb.odb_structure import iuphar_edges
from ebel.manager.rdbms.models import iuphar
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_interaction():
    """Get generic IUPHAR entry."""
    return _get_data(iuphar.IupharInteraction)


def get_ligandby_by_id():
    """Get IUPHAR ligand entry by ID."""
    ligand_id = request.args.get("id")
    return RDBMS.get_session().query(iuphar.IupharLigand).get(ligand_id).as_dict()


def get_interaction_by_target_uniprot():
    """Get IUPHAR interaction entry by target UniProt ID."""
    return get_interaction()


def get_interaction_by_target_gene_symbol():
    """Get IUPHAR interaction entry by target gene symbol."""
    return get_interaction()


def get_ebel_relation():
    """Get Iuphar related eBEL relations."""
    as_iuphar_edge = [x for x in iuphar_edges if x.name == "iuphar_interaction"][0]
    conf = {x.prop_name: x.prop_name for x in as_iuphar_edge.props}
    conf.update(
        {
            "edge_id": "@rid.asString()",
            "interactor_a_rid": "out.@rid.asString()",
            "interactor_a_name": "out.name",
            "interactor_a_namespace": "out.namespace",
            "interactor_b_rid": "in.@rid.asString()",
            "interactor_b_namespace": "in.namespace",
            "interactor_b_name": "in.name",
            "relation": "@class.asString()",
        }
    )
    sql = "SELECT "
    sql += ", ".join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM iuphar_interaction"

    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}
    pmid = request.args.get("pmid")
    if paras or pmid:
        wheres = []
        if paras:
            wheres += [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]
        if pmid:
            wheres += [f"{pmid} in pmids"]

        sql += " WHERE " + " AND ".join(wheres)

    return _get_paginated_ebel_query_result(sql)
